"""Helper functions for datapath configuration UI elements."""

import streamlit as st

from adijif.converters import converter as conv


def gen_datapath(converter: conv) -> int:
    """Generate datapath configuration UI elements for Streamlit.

    Args:
        converter: The converter object to generate the datapath for.

    Returns:
        The selected decimation or interpolation value.

    Raises:
        Exception: If no interpolation setting is found for DACs.
    """
    if converter.converter_type.lower() == "adc":
        decimation = 1
        if converter.datapath:
            if hasattr(converter.datapath, "cddc_decimations_available"):
                options = converter.datapath.cddc_decimations_available
                cddc_decimation = st.selectbox(
                    "CDDC Decimation",
                    options=options,
                    format_func=lambda x: str(x),
                )
                decimation = cddc_decimation
                v = len(converter.datapath.cddc_decimations)
                converter.datapath.cddc_decimations = [cddc_decimation] * v
            if hasattr(converter.datapath, "fddc_decimations_available"):
                options = converter.datapath.fddc_decimations_available
                fddc_decimation = st.selectbox(
                    "FDDC Decimation",
                    options=options,
                    format_func=lambda x: str(x),
                )
                decimation *= fddc_decimation
                v = len(converter.datapath.fddc_decimations)
                converter.datapath.fddc_decimations = [fddc_decimation] * v
        elif hasattr(converter, "decimation_available"):
            decimation = st.selectbox(
                "Decimation",
                options=converter.decimation_available,
                format_func=lambda x: str(x),
            )
            decimation = int(decimation)
            converter.decimation = decimation
    else:
        interpolation = 1
        if converter.datapath:
            if hasattr(converter.datapath, "cduc_interpolations_available"):
                options = converter.datapath.cduc_interpolations_available
                cduc_interpolation = st.selectbox(
                    "CDUC Interpolation",
                    options=options,
                    format_func=lambda x: str(x),
                )
                interpolation = cduc_interpolation
                converter.datapath.cduc_interpolation = cduc_interpolation
            if hasattr(converter.datapath, "fduc_interpolations_available"):
                options = converter.datapath.fduc_interpolations_available
                fduc_interpolation = st.selectbox(
                    "FDUC Interpolation",
                    options=options,
                    format_func=lambda x: str(x),
                )
                interpolation *= fduc_interpolation
                converter.datapath.fduc_interpolation = fduc_interpolation
        elif hasattr(converter, "interpolation_available"):
            interpolation = st.selectbox(
                "Interpolation",
                options=converter.interpolation_available,
                format_func=lambda x: str(x),
            )
            interpolation = int(interpolation)
            converter.interpolation = interpolation
        else:
            raise Exception("No interpolation setting found")

        decimation = interpolation

    return decimation



"""JESD parameterization definitions and helper functions."""

from abc import ABCMeta, abstractmethod
from typing import Dict, List, Union

from adijif.solvers import CpoSolveResult


class jesd(metaclass=ABCMeta):
    """JESD interface class to manage JESD notations and definitions."""

    # Lane rate min/max defaulting to JESD spec (parts may differ)
    bit_clock_min_available = {"jesd204b": 312.5e6, "jesd204c": 312.5e6}
    bit_clock_max_available = {"jesd204b": 12.5e9, "jesd204c": 32e9}

    solver = "CPLEX"

    _parameters_to_return = [
        "bit_clock",
        "multiframe_clock",
        "sample_clock",
        "F",
        "HD",
        "K",
        "L",
        "M",
        "Np",
        "S",
        "CS",
        "jesd_class",
        "converter_clock",
    ]

    _skip_clock_validation = False

    def __init__(self, sample_clock: int, M: int, L: int, Np: int, K: int) -> None:
        """Initialize JESD device through link parameterization.

        Args:
            sample_clock (int): Human readable string describing the exception.
            M (int): Number of virtual converters
            L (int): Number of lanes
            Np (int): Number of bits per sample
            K (int): Frames per multiframe

        """
        self.sample_clock = sample_clock
        self.K = K
        self.L = L
        self.M = M
        self.Np = Np
        # self.S = S

    def _check_clock_relations(self) -> None:
        """Check clock relations between clocks and JESD parameters."""
        sc = self.sample_clock
        assert sc == self.frame_clock * self.S, "sample_clock != S * frame_clock"
        if self.jesd_class == "jesd204b":
            assert sc == (self.bit_clock / 10 / self.F) * self.S
            assert sc == (self.multiframe_clock * self.K * self.S)

    def get_jesd_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect JESD related parameters, includes modes and clocks.

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of JESD parameters
        """
        if solution:  # type: ignore
            self.solution = solution
        cfg = {p: getattr(self, p) for p in self._parameters_to_return}
        cfg["jesd_mode"] = self._check_valid_jesd_mode()
        return cfg

    def validate_clocks(self) -> None:
        """Validate all clocks clock settings are within range."""
        if not self._skip_clock_validation:
            for name in ["bit", "sample"]:
                clk = getattr(self, name + "_clock")
                lim = getattr(self, name + "_clock_max")
                assert clk <= lim, (
                    name + f" clock too fast for device {clk} (limit: {lim})"
                )
                lim = getattr(self, name + "_clock_min")
                assert clk >= lim, (
                    name + f" clock too slow for device {clk} (limit: {lim})"
                )

    @property
    def bit_clock_min(self) -> Union[int, float]:
        """Get bit clock (lane rate) minimum based on JESD mode.

        Returns:
            int: bit clock in bits per second
        """
        if isinstance(self.jesd_class, list):
            return self.bit_clock_min_available["jesd204b"]
        return self.bit_clock_min_available[self.jesd_class]

    @property
    def bit_clock_max(self) -> Union[int, float]:
        """Get bit clock (lane rate) maximum based on JESD mode.

        Returns:
            int: bit clock in bits per second
        """
        if isinstance(self.jesd_class, list):
            return self.bit_clock_max_available["jesd204b"]
        return self.bit_clock_max_available[self.jesd_class]

    _jesd_class = "jesd204b"

    @property
    def jesd_class(self) -> Union[str, List[str]]:
        """Get JESD selected mode. Wil be either jesd204b or jesd204c."""
        return self._jesd_class

    @jesd_class.setter
    def jesd_class(self, value: str) -> None:
        """Set JESD selected mode and must be either jesd204b or jesd204c.

        Args:
            value (str): String of JESD class and must be jesd204b or jesd204c

        Raises:
            Exception: Invalid JESD class selected
        """
        if value not in self.available_jesd_modes:
            raise Exception(
                f"Invalid JESD class. Valid are: {self.available_jesd_modes}"
            )
        self._jesd_class = value
        if value == "jesd204b":
            self._encoding = "8b10b"
        else:
            self._encoding = "64b66b"

    def _check_jesd_config(self) -> None:
        """Check if bit clock is within JESD limits based on supported standard.

        Raises:
            Exception: bit clock (lane rate) too high for JESD mode or invalid
        """
        if "jesd204c" in self.available_jesd_modes:
            if self.bit_clock > 32e9:
                raise Exception(
                    f"bit clock (lane rate) {self.bit_clock} too high for JESD204C"
                )
        elif "jesd204b" in self.available_jesd_modes:
            if self.bit_clock > 12.5e9:
                raise Exception(
                    f"bit clock (lane rate) {self.bit_clock} too high for JESD204B"
                )
        else:
            raise Exception(f"JESD mode(s) {self.available_jesd_modes}")

    @property
    @abstractmethod
    def available_jesd_modes(self) -> List[str]:
        """Available JESD modes supported by device.

        Must be a list of strings

        Raises:
            NotImplementedError: If child classes do not implement method/property
        """
        raise NotImplementedError  # pragma: no cover

    """ CS: Control bits per conversion sample 0-3"""
    _CS = 0
    CS_available = [0, 1, 2, 3]

    @property
    def CS(self) -> Union[int, float]:
        """Get Control bits per conversion sample.

        Returns:
            int: Control bits per conversion sample
        """
        return self._CS

    @CS.setter
    def CS(self, value: int) -> None:
        """Set Control bits per conversion sample.

        Args:
            value (int): Control bits per conversion sample

        Raises:
            Exception: CS not an integer or not in range
        """
        if int(value) != value:
            raise Exception("CS must be an integer")
        if value not in self.CS_available:
            raise Exception("CS not in range for device")
        self._CS = value

    """ CF: Control word per frame clock period per link 0-32 """
    _CF = 0
    CF_available = [0, 1]

    @property
    def CF(self) -> Union[int, float]:
        """Get Control words per frame clock period per link.

        Returns:
            int: Control words per frame clock period per link
        """
        return self._CF

    @CF.setter
    def CF(self, value: int) -> None:
        """Set Control words per frame clock period per link.

        Args:
            value (int): Control words per frame clock period per link

        Raises:
            Exception: CF not an integer or not in range
        """
        if int(value) != value:
            raise Exception("CF must be an integer")
        if value not in self.CF_available:
            raise Exception("CF not in range for device")
        self._CF = value

    # Encoding functions

    encodings_n = {"8b10b": 8, "64b66b": 64}
    encodings_d = {"8b10b": 10, "64b66b": 66}
    _encoding = "8b10b"

    @property
    def encoding(self) -> str:
        """Get JESD FEC encoding.

        Current options are: "8b10b", "64b66b"

        Returns:
            str: String of supported encodings.
        """
        return self._encoding

    @encoding.setter
    def encoding(self, value: str) -> None:
        """Set JESD FEC encoding.

        Current options are: "8b10b", "64b66b"

        Args:
            value (str): String of desired encoding to use

        Raises:
            Exception: If encoding selected that is not supported
        """
        if self._check_encoding(value):
            raise Exception(
                "JESD encoding not possible due to available modes: {}".format(
                    self.available_jesd_modes
                )
            )
        self._encoding = value

    @property
    def encoding_d(self) -> Union[int, float]:
        """Get JESD FEC encoding denominator.

        Current options are: 10 or 66

        Returns:
            int: Denominator of link encoding.
        """
        return self.encodings_d[self._encoding]

    @property
    def encoding_n(self) -> Union[int, float]:
        """Get JESD FEC encoding numerator.

        Current options are: 8 or 64

        Returns:
            int: Numerator of link encoding.
        """
        return self.encodings_n[self._encoding]

    def _check_encoding(self, encode: str) -> bool:
        if "jesd204c" in self.available_jesd_modes:
            allowed_encodings = ["8b10b", "64b66b"]
        else:
            allowed_encodings = ["8b10b"]
        return encode not in allowed_encodings

    # SCALERS
    @property
    @abstractmethod
    def K_available(self) -> List[int]:
        """Allowable K settings for device.

        Must be a list ints

        Raises:
            NotImplementedError: If child classes do not implement method/property
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def L_available(self) -> List[int]:
        """Allowable L settings for device.

        Must be a list ints

        Raises:
            NotImplementedError: If child classes do not implement method/property
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def M_available(self) -> List[int]:
        """Allowable M settings for device.

        Must be a list ints

        Raises:
            NotImplementedError: If child classes do not implement method/property
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def N_available(self) -> List[int]:
        """Allowable N settings for device.

        Must be a list ints

        Raises:
            NotImplementedError: If child classes do not implement method/property
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def Np_available(self) -> List[int]:
        """Allowable Np settings for device.

        Must be a list ints

        Raises:
            NotImplementedError: If child classes do not implement method/property
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def F_available(self) -> List[int]:
        """Allowable F settings for device.

        Must be a list ints

        Raises:
            NotImplementedError: If child classes do not implement method/property
        """
        raise NotImplementedError  # pragma: no cover

    """ bits
        Usually:
            32 for JESD204B
            64 for JESD204C
    """
    _data_path_width = 32

    @property
    def data_path_width(self) -> Union[int, float]:
        """Get JESD data path width in bits.

        Current options are: 32 (204B) and 64 (204C)

        Returns:
            int: Numerator of link encoding.
        """
        return self._data_path_width

    @data_path_width.setter
    def data_path_width(self, value: int) -> None:
        """Set JESD data path width in bits.

        Current options are: 32 (204B) and 64 (204C)

        Args:
            value (int): Data path width in bits

        Raises:
            Exception: If DMA width is not an integer
        """
        if int(value) != value:
            raise Exception("data_path_width must be an integer")
        self._data_path_width = value

    """ HD: High-density mode (Single sample split over multiple lanes)"""
    # HD_min = 0
    # HD_max = 1
    _HD = 0
    HD_available = [0, 1]

    @property
    def HD(self) -> Union[int, float]:
        """Get High density mode.

        Returns:
            int: High density mode
        """
        return self._HD

    @HD.setter
    def HD(self, value: int) -> None:
        """Set High density mode.

        Args:
            value (int): High density mode

        Raises:
            Exception: HD not an integer or not in range
        """
        if int(value) != value:
            raise Exception("HD must be an integer")
        if value not in self.HD_available:
            raise Exception("HD not in range for device")
        self._HD = value

    """ K: Frames per multiframe
        17/F <= K <= 32
    """
    # K_min = 4
    # K_max = 32
    # K_available = [4, 8, 12, 16, 20, 24, 28, 32]
    _K = 4

    @property
    def K(self) -> Union[int, float]:
        """Get Frames per multiframe.

        17/F <= K <= 32, is generally a multiple of 2

        Returns:
            int: Number of frames per multiframe
        """
        return self._K

    @K.setter
    def K(self, value: int) -> None:
        """Set Frames per multiframe.

        Args:
            value (int): Frames per multiframe

        Raises:
            Exception: K not an integer or not in range
        """
        if int(value) != value:
            raise Exception("K must be an integer")
        if value not in self.K_available:
            raise Exception("K not in range for device")
        self._K = value

    @property
    def D(self) -> Union[int, float]:
        """FIXME."""
        return self._data_path_width * self.encoding_d / self.encoding_n

    """ S: Samples per converter per frame"""
    _S = 1

    @property
    def S(self) -> Union[int, float]:
        """Get Samples per converter per frame.

        S == F/(M*Np) * encoding_p * L

        Returns:
            int: Samples per converter per frame
        """
        return self._S

    @S.setter
    def S(self, value: int) -> None:
        """Set samples per converter per frame.

        Args:
            value (int): Samples per converter per frame

        Raises:
            Exception: S not an integer or not in range
        """
        if int(value) != value:
            raise Exception("S must be an integer")
        if value not in self.S_available:
            raise Exception(f"S not in range for device. Got {value}")
        self._S = value

    """ L: Lanes per link """
    # L_min = 1
    # L_max = 8
    # L_available = [1, 2, 4, 8]
    _L = 1

    @property
    def L(self) -> Union[int, float]:
        """Get lanes per link.

        Generally a multiple of 2

        Returns:
            int: Number of frames per multiframe
        """
        return self._L

    @L.setter
    def L(self, value: int) -> None:
        """Set lanes per link.

        Args:
            value (int): Lanes per link

        Raises:
            Exception: L not an integer or not in range
        """
        if int(value) != value:
            raise Exception("L must be an integer")
        if value not in self.L_available:
            raise Exception("L not in range for device")
        self._L = value

    """ M: Number of virtual converters """
    # M_min = 1
    # M_max = 8
    # M_available = [1, 2, 4, 8, 16, 32]
    _M = 1

    @property
    def M(self) -> Union[int, float]:
        """Get number of virtual converters.

        Generally a power of 2

        Returns:
            int: Number of frames per multiframe
        """
        return self._M

    @M.setter
    def M(self, value: int) -> None:
        """Set number of virtual converters.

        Args:
            value (int): Number of virtual converters

        Raises:
            Exception: M not an integer or not in range
        """
        if int(value) != value:
            raise Exception("M must be an integer")
        if value not in self.M_available:
            raise Exception("M not in range for device")
        self._M = value

    """ N: Number of non-dummy bits per sample """
    # N_min = 12
    # N_max = 16
    # N_available = [12, 14, 16]
    _N = 12

    @property
    def N(self) -> Union[int, float]:
        """Get number of non-dummy bits per sample.

        Generally a multiple of 2

        Returns:
            int: Number of non-dummy bits per sample
        """
        return self._N

    @N.setter
    def N(self, value: int) -> None:
        """Set number of non-dummy bits per sample.

        Args:
            value (int): Number of non-dummy bits per sample

        Raises:
            Exception: N not an integer or not in range
        """
        if int(value) != value:
            raise Exception("N must be an integer")
        if value not in self.N_available:
            raise Exception("N not in range for device")
        self._N = value

    """ Np: Number of bits per sample """
    # Np_min = 12
    # Np_max = 16
    # Np_available = [12, 14, 16]
    _Np = 16

    @property
    def Np(self) -> Union[int, float]:
        """Get number of bits per sample.

        Generally a multiple of 2

        Returns:
            int: Number of bits per sample
        """
        return self._Np

    @Np.setter
    def Np(self, value: int) -> None:
        """Set number of bits per sample.

        Args:
            value (int): Number of bits per sample

        Raises:
            Exception: Np not an integer or not in range
        """
        if int(value) != value:
            raise Exception("Np must be an integer")
        if value not in self.Np_available:
            msg = "Np not in range for device. "
            msg += f"Got {value}, valid are {self.Np_available}"
            raise Exception(msg)
        self._Np = value

    # DERIVED SCALERS
    """ F: Octets per frame per link
        This is read-only since it depends on L,M,Np,S, and encoding
    """
    # F_min = 1
    # F_max = 16
    # F_available = [1, 2, 4, 8, 16]
    _F = 1

    @property
    def F(self) -> Union[int, float]:
        """Get octets per frame per link.

        Generally a power of 2

        Returns:
            int: Number of octets per frame per link
        """
        return self._F

    @F.setter
    def F(self, value: int) -> None:
        """Set octets per frame per link.

        Args:
            value (int): Number of octets per frame per link

        Raises:
            Exception: F not an integer or not in range
        """
        if int(value) != value:
            raise Exception("F must be an integer")
        if value not in self.F_available:
            raise Exception("F not in range for device")
        self._F = value

    # CLOCKS
    """ sample_clock: Data rate after decimation stages in Samples/second """

    _sample_clock = 122.88e6

    @property
    def sample_clock(self) -> Union[int, float]:
        """Data rate after decimation stages in Samples/second.

        Returns:
            int: Data rate in samples per second
        """
        return self._sample_clock

    @sample_clock.setter
    def sample_clock(self, value: int) -> None:
        """Data rate after decimation stages in Samples/second.

        Args:
            value (int): Number of octets per frame per link
        """
        self._sample_clock = value

    @property
    def frame_clock(self) -> Union[int, float]:
        """frame_clock in frames per second.

        frame_clock == sample_clock / S

        Returns:
            int: Data rate in samples per second
        """
        return self.sample_clock / self.S

    @property
    def multiframe_clock(self) -> Union[int, float]:
        """multiframe_clock: aka LMFC in frames per multiframe.

        multiframe_clock == frame_clock / K

        Returns:
            int: Frames per multiframe
        """
        return self.frame_clock / self.K

    @property
    def bit_clock(self) -> Union[int, float]:
        """bit_clock: aka line rate aka lane rate.

        bit_clock == (M * S * Np * encoding_d/encoding_n * frame_clock) / L

        Returns:
            int: Bits per second aka lane rate
        """
        return (
            (self.M / self.L)
            * self.Np
            * (self.encoding_d / self.encoding_n)
            * self.sample_clock
        )

    @bit_clock.setter
    def bit_clock(self, value: int) -> None:
        """bit_clock: aka line rate aka lane rate.

        bit_clock == (M * S * Np * encoding_d/encoding_n * frame_clock) / L

        Args:
            value (int): Bits per second aka lane rate
        """
        # This actually sets sample_clock
        # frame_clock = bit_clock*L*encoding_n/encoding_d / (M*S*Np)
        # sample_clock = bit_clock*L*encoding_n/encoding_d / (M*Np)
        value_cal = (
            value * self.L * self.encoding_n / self.encoding_d / (self.M * self.Np)
        )
        self._sample_clock = value_cal

    @property
    def device_clock(self) -> Union[int, float]:
        """device_clock is the lane rate over D.

        device_clock == bit_clock / D

        Returns:
            int: bits per second per device
        """
        return self.bit_clock / self.D



"""FPGA parent metaclass to maintain consistency for all FPGA classes."""

from abc import ABCMeta, abstractmethod
from typing import Dict, List, Optional, Union

from adijif.common import core
from adijif.converters.converter import converter as conv
from adijif.gekko_trans import gekko_translation
from adijif.solvers import CpoSolveResult


class fpga(core, gekko_translation, metaclass=ABCMeta):
    """Parent metaclass for all FPGA classes."""

    """Generate another clock for link layer output.
    This is used when the GEARBOX is enabled in HDL within link layer core.
    """
    requires_separate_link_layer_out_clock: bool = False

    @property
    @abstractmethod
    def determine_cpll(self) -> None:
        """Try to use CPLL for clocking.

        This method is only used in brute-force classes

        Raises:
            NotImplementedError: Method not implemented
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def determine_qpll(self) -> None:
        """Try to use QPLL for clocking.

        This method is only used in brute-force classes

        Raises:
            NotImplementedError: Method not implemented
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def get_config(  # type: ignore
        self,
        converter: conv,
        fpga_ref: Union[float, int],
        solution: Optional[CpoSolveResult] = None,
    ) -> Union[List[Dict], Dict]:
        """Extract clocking configuration from solutions.

        Args:
            converter (conv): Converter object connected to FPGA who config is
                collected
            fpga_ref (int or float): Reference clock generated for FPGA for specific
                converter
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Raises:
            NotImplementedError: Method not implemented
        """
        raise NotImplementedError  # pragma: no cover



"""Helper functions for JESD mode handling."""

import copy
import logging
from typing import Any, Dict, List, Optional, Tuple

from adijif.utils import get_jesd_mode_from_params

log = logging.getLogger(__name__)


def get_jesd_controls(converter: object) -> Tuple[Dict[str, List[Any]], Dict[str, Any]]:
    """Get JESD control options from converter.

    Args:
        converter: JESD converter object

    Returns:
        Tuple of (options dict, all_modes dict)

    Raises:
        Exception: When mode settings are not consistent across all subclasses
    """
    options_to_skip = ["global_index", "decimations", "interpolations"]

    # Derive configurable jesd settings for specific converter based on mode table
    all_modes = converter.quick_configuration_modes
    all_modes = copy.deepcopy(all_modes)
    subclasses = list(all_modes.keys())

    mode_knobs = None

    for subclass in subclasses:
        # Pick first mode of subclass and extract settings
        ks = all_modes[subclass].keys()
        ks = list(ks)
        if len(ks) == 0:
            # print(f"No modes found for subclass {subclass}")
            continue
        first_mode = list(all_modes[subclass].keys())[0]
        mode_settings = all_modes[subclass][first_mode].keys()
        mode_settings = sorted(mode_settings)
        if not mode_knobs:
            mode_knobs = mode_settings
            continue

        # Compare settings in other subclasses to make sure they are the same
        if mode_settings != mode_knobs:
            differences = set(mode_settings) ^ set(mode_knobs)
            log.warning(
                f"Mode settings are not consistent across all subclasses: {differences}"
            )
            raise Exception("Mode settings are not consistent across all subclasses")

    # Parse all options for each control across modes
    options = {}
    for setting in mode_knobs:
        if setting in options_to_skip:
            continue
        options[setting] = []
        for subclass in subclasses:
            for mode in all_modes[subclass]:
                data = all_modes[subclass][mode][setting]
                if isinstance(data, list):
                    print(f"Skipping list data for {setting}")
                    continue
                options[setting].append(data)

    # Make sure options only contain unique values
    for option in options:
        options[option] = list(set(options[option]))

    return options, all_modes


def get_valid_jesd_modes(
    converter: object, all_modes: Dict[str, Any], selections: Dict[str, Any]
) -> Tuple[List[Dict[str, Any]], Optional[List[Dict[str, Any]]]]:
    """Get valid JESD modes based on selections.

    Args:
        converter: JESD converter object
        all_modes: Dictionary of all available modes
        selections: User selections for filtering

    Returns:
        Tuple of (modes info list, found modes list or None)
    """
    modes_all_info: Any = {}

    try:
        found_modes = get_jesd_mode_from_params(converter, **selections)

    except Exception:
        log.info("No modes found")
        found_modes = None
        return modes_all_info, found_modes

    options_to_skip = ["global_index", "decimations"]

    # Get remaining mode parameters
    modes_all_info = []
    for mode in found_modes:
        jesd_cfg = all_modes[mode["jesd_class"]][mode["mode"]]
        jesd_cfg["mode"] = mode["mode"]
        jesd_cfg["jesd_class"] = mode["jesd_class"]

        # Remove options to skip
        for option in options_to_skip:
            jesd_cfg.pop(option, None)

        modes_all_info.append(jesd_cfg)

    # For each mode calculate the clocks and if valid
    for mode in modes_all_info:
        rate = converter.sample_clock
        # print("A", converter.sample_clock)
        converter.set_quick_configuration_mode(mode["mode"], mode["jesd_class"])
        # print("B", converter.sample_clock)
        log.warning("Logical BUG")
        converter.sample_clock = rate

        clocks = {
            "Sample Rate (MSPS)": converter.sample_clock / 1e6,
            "Lane Rate (GSPS)": converter.bit_clock / 1e9,
        }

        for clock in clocks:
            mode[clock] = clocks[clock]

        try:
            converter.validate_config()
            mode["Valid"] = "Yes"
        except Exception:
            mode["Valid"] = "No"

    # from pprint import pprint
    # pprint(modes_all_info)

    return modes_all_info, found_modes



"""AD9528 clock chip model."""

from typing import Dict, List, Union

from adijif.clocks.ad9528_bf import ad9528_bf
from adijif.solvers import CpoExpr, CpoSolveResult, GK_Intermediate


class ad9528(ad9528_bf):
    """AD9528 clock chip model.

    This model currently supports VCXO+PLL2 configurations
    """

    name = "AD9528"

    # Ranges
    """ VCO divider """
    m1_available = [3, 4, 5]
    """ Output dividers """
    d_available = [*range(1, 1024)]
    """ sysref dividers """
    k_available = [*range(0, 65536)]
    """ VCXO multiplier """
    n2_available = [*range(1, 256)]
    """ VCO calibration dividers """
    a_available = [0, 1, 2, 3]
    b_availble = [*range(3, 64)]
    # N = (PxB) + A, P=4, A==[0,1,2,3], B=[3..63]
    # See table 46 of DS for limits
    """ VCXO dividers """
    r1_available = [*range(1, 32)]

    # State management
    _clk_names: List[str] = []

    # Defaults
    _m1: Union[List[int], int] = [3, 4, 5]
    _d: Union[List[int], int] = [*range(1, 1024)]
    _k: Union[List[int], int] = k_available
    _n2: Union[List[int], int] = n2_available
    _r1: Union[List[int], int] = [*range(1, 32)]
    _a: Union[List[int], int] = [*range(0, 4)]
    _b: Union[List[int], int] = [*range(3, 64)]

    # Limits
    vco_min = 3450e6
    vco_max = 4025e6
    pfd_max = 275e6

    minimize_feedback_dividers = True

    use_vcxo_double = False
    vcxo = 125e6

    # sysref parameters
    sysref_external = False
    _sysref = None

    @property
    def m1(self) -> Union[int, List[int]]:
        """VCO divider path 1.

        Valid dividers are 3,4,5

        Returns:
            int: Current allowable dividers
        """
        return self._m1

    @m1.setter
    def m1(self, value: Union[int, List[int]]) -> None:
        """VCO divider path 1.

        Valid dividers are 3,4,5

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.m1_available, "m1")
        self._m1 = value

    @property
    def d(self) -> Union[int, List[int]]:
        """Output dividers.

        Valid dividers are 1->1023

        Returns:
            int: Current allowable dividers
        """
        return self._d

    @d.setter
    def d(self, value: Union[int, List[int]]) -> None:
        """Output dividers.

        Valid dividers are 1->1023

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.d_available, "d")
        self._d = value

    @property
    def k(self) -> Union[int, List[int]]:
        """Sysref dividers.

        Valid dividers are 0->65535

        Returns:
            int: Current allowable dividers
        """
        return self._k

    @k.setter
    def k(self, value: Union[int, List[int]]) -> None:
        """Sysref dividers.

        Valid dividers are 0->65535

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.d_available, "k")
        self._k = value

    @property
    def n2(self) -> Union[int, List[int]]:
        """n2: VCO feedback divider.

        Valid dividers are 1->255

        Returns:
            int: Current allowable dividers
        """
        return self._n2

    @n2.setter
    def n2(self, value: Union[int, List[int]]) -> None:
        """VCO feedback divider.

        Valid dividers are 1->255

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.n2_available, "n2")
        self._n2 = value

    @property
    def r1(self) -> Union[int, List[int]]:
        """VCXO input dividers.

        Valid dividers are 1->31

        Returns:
            int: Current allowable dividers
        """
        return self._r1

    @r1.setter
    def r1(self, value: Union[int, List[int]]) -> None:
        """VCXO input dividers.

        Valid dividers are 1->31

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.r1_available, "r1")
        self._r1 = value

    @property
    def a(self) -> Union[int, List[int]]:
        """VCO calibration divider 1.

        Valid dividers are 0->3

        Returns:
            int: Current allowable dividers
        """
        return self._a

    @a.setter
    def a(self, value: Union[int, List[int]]) -> None:
        """VCO calibration divider 1.

        Valid dividers are 0->3

        Args:
            value (int, list[int]): Allowable values for counter

        """
        self._check_in_range(value, self.a_available, "a")
        self._a = value

    @property
    def b(self) -> Union[int, List[int]]:
        """VCO calibration divider 2.

        Valid dividers are 3->63

        Returns:
            int: Current allowable dividers
        """
        return self._b

    @b.setter
    def b(self, value: Union[int, List[int]]) -> None:
        """VCO calibration divider 2.

        Valid dividers are 3->63

        Args:
            value (int, list[int]): Allowable values for counter

        """
        self._check_in_range(value, self.b_available, "b")
        self._b = value

    @property
    def vco(self) -> float:
        """VCO Frequency in Hz.

        Returns:
            float: computed VCO frequency
        """
        r1 = self._get_val(self.config["r1"])
        m1 = self._get_val(self.config["m1"])
        n2 = self._get_val(self.config["n2"])

        return self.vcxo / r1 * m1 * n2

    @property
    def sysref(self) -> int:
        """SYSREF Frequency in Hz.

        Returns:
            int: computed sysref frequency
        """
        r1 = self._get_val(self.config["r1"])
        k = self._get_val(self.config["k"])

        if self.sysref_external:
            sysref_src = self.vcxo
        else:
            sysref_src = self.vcxo / r1

        return sysref_src / (2 * k)

    @sysref.setter
    def sysref(self, value: Union[int, float]) -> None:
        """Set sysref frequency.

        Args:
            value (int, float): Frequency
        """
        self._sysref = int(value)

    def get_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect internal clock chip configuration and output clock definitions
        leading to connected devices (converters, FPGAs)

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration

        Raises:
            Exception: If solver is not called first
        """
        if not self._clk_names:
            raise Exception("set_requested_clocks must be called before get_config")

        if solution:
            self.solution = solution

        # out_dividers = [solution.get_value(x) for x in self.config["out_dividers"]]
        out_dividers = [self._get_val(x) for x in self.config["out_dividers"]]

        # Handle different vcxo types
        if hasattr(self, "vcxo_arb") and self.vcxo_arb:
            # arb_source case
            vcxo_cfg = self.vcxo_arb.get_config(self.solution)  # type: ignore
            vcxo_val = list(vcxo_cfg.values())[0]
        else:
            # int/float or range case
            vcxo_val = self._get_val(self.vcxo)

        config: Dict = {
            "vcxo": vcxo_val / 2 if self.use_vcxo_double else vcxo_val,
            "vco": self.vco,
            "r1": self._get_val(self.config["r1"]),
            "n2": self._get_val(self.config["n2"]),
            "m1": self._get_val(self.config["m1"]),
            "a": self._get_val(self.config["a"]),
            "b": self._get_val(self.config["b"]),
            "out_dividers": out_dividers,
            "output_clocks": [],
        }

        if self._sysref:
            config["k"] = self._get_val(self.config["k"])
            config["sysref"] = self.sysref

        clk = self.vcxo * config["n2"] / config["r1"]

        output_cfg = {}
        for i, div in enumerate(out_dividers):
            rate = clk / div
            output_cfg[self._clk_names[i]] = {
                "rate": rate,
                "divider": div,
            }

        config["output_clocks"] = output_cfg

        self._saved_solution = config

        return config

    def _setup_solver_constraints(self, vcxo: int) -> None:
        """Apply constraints to solver model.

        Args:
            vcxo (int): VCXO frequency in hertz

        Raises:
            Exception: Unknown solver
        """
        if not isinstance(vcxo, (float, int)):
            vcxo_result = vcxo(self.model)
            # Handle range type (returns dict with "range" key)
            if isinstance(vcxo_result, dict):
                self.vcxo = vcxo_result["range"]
                self.vcxo_arb = None
            # Handle arb_source type (returns direct expression)
            else:
                self.vcxo = vcxo_result
                self.vcxo_arb = vcxo  # Store original for get_config
        else:
            self.vcxo = vcxo
            self.vcxo_arb = None

        self.config = {
            "r1": self._convert_input(self._r1, "r1"),
            "m1": self._convert_input(self._m1, "m1"),
            "k": self._convert_input(self._k, "k"),
            "n2": self._convert_input(self._n2, "n2"),
            "a": self._convert_input(self._a, "a"),
            "b": self._convert_input(self._b, "b"),
        }
        # self.config = {"r1": self.model.Var(integer=True, lb=1, ub=31, value=1)}
        # self.config["m1"] = self.model.Var(integer=True, lb=3, ub=5, value=3)
        # self.config["n2"] = self.model.Var(integer=True, lb=12, ub=255, value=12)

        # PLL2 equations
        self._add_equation(
            [
                self.vcxo / self.config["r1"] <= self.pfd_max,
                self.vcxo / self.config["r1"] * self.config["m1"] * self.config["n2"]
                <= self.vco_max,
                self.vcxo / self.config["r1"] * self.config["m1"] * self.config["n2"]
                >= self.vco_min,
                4 * self.config["b"] + self.config["a"] >= 16,
                4 * self.config["b"] + self.config["a"]
                == self.config["m1"] * self.config["n2"],
            ]
        )
        # Objectives
        if self.minimize_feedback_dividers:
            if self.solver == "CPLEX":
                self._add_objective(self.config["n2"])
                # self.model.minimize(self.config["n2"])
            elif self.solver == "gekko":
                self.model.Obj(self.config["n2"])
            else:
                raise Exception("Unknown solver {}".format(self.solver))
        # self.model.Obj(self.config["n2"] * self.config["m1"])

    def _setup(self, vcxo: int) -> None:
        # Setup clock chip internal constraints

        # FIXME: ADD SPLIT m1 configuration support

        # Setup clock chip internal constraints
        if self.use_vcxo_double:
            vcxo *= 2
        self._setup_solver_constraints(vcxo)

        # Add requested clocks to output constraints
        self.config["out_dividers"] = []
        self._clk_names = []  # Reset clock names

    def _get_clock_constraint(
        self, clk_name: str
    ) -> Union[int, float, CpoExpr, GK_Intermediate]:
        """Get abstract clock output.

        Args:
            clk_name (str):  String of clock name

        Returns:
            (int or float or CpoExpr or GK_Intermediate): Abstract
                or concrete clock reference
        """
        od = self._convert_input(self._d, "d_" + str(clk_name))
        self.config["out_dividers"].append(od)
        self._clk_names.append(clk_name)
        return self.vcxo / self.config["r1"] * self.config["n2"] / od

    def set_requested_clocks(
        self,
        vcxo: int,
        out_freqs: List,
        clk_names: List[str],
    ) -> None:
        """Define necessary clocks to be generated in model.

        Args:
            vcxo (int): VCXO frequency in hertz
            out_freqs (List): list of required clocks to be output
            clk_names (List[str]):  list of strings of clock names

        Raises:
            Exception: If len(out_freqs) != len(clk_names)
        """
        if len(clk_names) != len(out_freqs):
            raise Exception("clk_names is not the same size as out_freqs")

        # Setup clock chip internal constraints
        self._setup(vcxo)
        self._clk_names = clk_names

        if self._sysref:
            if self.sysref_external:
                sysref_src = self.vcxo
            else:
                sysref_src = self.vcxo / self.config["r1"]

            self._add_equation([sysref_src / (2 * self.config["k"]) == self._sysref])

        # Add requested clocks to output constraints
        for out_freq, name in zip(out_freqs, clk_names):  # noqa: B905
            # od = self.model.Var(integer=True, lb=1, ub=256, value=1)
            od = self._convert_input(self._d, f"d_{name}_{out_freq}")
            # od = self.model.sos1([n*n for n in range(1,9)])
            self._add_equation(
                [self.vcxo / self.config["r1"] * self.config["n2"] / od == out_freq]
            )
            self.config["out_dividers"].append(od)



"""ADRV9009 transceiver clocking model."""

from abc import ABCMeta
from typing import Any, Dict, List, Union

import numpy as np

from adijif.converters.adrv9009_bf import adrv9009_bf

from ..solvers import CpoModel  # type: ignore # noqa: I202,BLK100
from ..solvers import GEKKO, CpoSolveResult
from .adc import adc
from .adrv9009_util import quick_configuration_modes_rx  # type: ignore
from .adrv9009_util import _extra_jesd_check, quick_configuration_modes_tx
from .converter import converter
from .dac import dac

# References
# https://ez.analog.com/wide-band-rf-transceivers/design-support-adrv9008-1-adrv9008-2-adrv9009/f/q-a/103757/adrv9009-clock-configuration/308013#308013


class adrv9009_core(converter, metaclass=ABCMeta):
    """ADRV9009 transceiver clocking model.

    This model manage the JESD configuration and input clock constraints.
    External LO constraints are not modeled.

    Clocking: ADRV9009 uses onboard PLLs to generate the JESD clocks

        Lane Rate = I/Q Sample Rate * M * Np * (10 / 8) / L
        Lane Rate = sample_clock * M * Np * (10 / 8) / L
    """

    device_clock_available = None  # FIXME
    device_clock_ranges = None  # FIXME

    name = "ADRV9009"

    # JESD configurations
    quick_configuration_modes = None  # FIXME
    available_jesd_modes = ["jesd204b"]
    M_available = [1, 2, 4]
    L_available = [1, 2, 3, 4, 6, 8]
    N_available = [12, 16]
    Np_available = [12, 16, 24]
    F_available = [1, 2, 3, 4, 8]
    S_available = [1]  # FIXME?
    K_available = [*np.arange(1, 32 + 1)]
    CS_available = [0]
    CF_available = [0]

    # Clock constraints
    converter_clock_min = 39.063e6 * 8
    converter_clock_max = 12288e6

    sample_clock_min = 39.063e6
    sample_clock_max = 491520000

    device_clock_min = 10e6
    device_clock_max = 1e9

    clocking_option_available = ["integrated_pll"]
    _clocking_option = "integrated_pll"

    # Divider ranges
    input_clock_dividers_available = [1 / 2, 1, 2, 4, 8, 16]
    input_clock_dividers_times2_available = [1, 2, 4, 8, 16, 32]

    _lmfc_divisor_sysref_available = [*range(1, 20)]

    # Unused
    max_rx_sample_clock = 250e6
    max_tx_sample_clock = 500e6
    max_obs_sample_clock = 500e6

    def _check_valid_internal_configuration(self) -> None:
        # FIXME
        pass

    def get_required_clock_names(self) -> List[str]:
        """Get list of strings of names of requested clocks.

        This list of names is for the clocks defined by
        get_required_clocks

        Returns:
            List[str]: List of strings of clock names mapped by get_required_clocks
        """
        return ["adrv9009_ref_clk", "adrv9009_sysref"]

    def get_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect internal converter configuration and output clock definitions
        leading to connected devices (clock chips, FPGAs)

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration
        """
        if solution:
            self.solution = solution
        return {}


class adrv9009_clock_common(adrv9009_core, adrv9009_bf):
    """ADRV9009 class managing common singleton (Rx,Tx) methods."""

    def _check_valid_jesd_mode(self) -> None:
        """Verify current JESD configuration for part is valid."""
        _extra_jesd_check(self)
        return super()._check_valid_jesd_mode()

    def get_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect internal converter configuration and output clock definitions
        leading to connected devices (clock chips, FPGAs)

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration
        """
        return {"clocking_option": self.clocking_option}

    def _gekko_get_required_clocks(self) -> List[Dict]:
        possible_sysrefs = []
        for n in range(1, 20):
            r = self.multiframe_clock / (n * n)
            if r == int(r):
                possible_sysrefs.append(r)

        self.config = {"sysref": self.model.sos1(possible_sysrefs)}
        self.model.Obj(self.config["sysref"])

        possible_device_clocks = []
        for div in self.input_clock_dividers_available:
            dev_clock = self.sample_clock / div
            if self.device_clock_min <= dev_clock <= self.device_clock_max:
                possible_device_clocks.append(dev_clock)

        self.config["ref_clk"] = self.model.sos1(possible_device_clocks)
        # self.model.Obj(-1 * self.config["device_clock"])

        return [self.config["ref_clk"], self.config["sysref"]]

    def get_required_clocks(self) -> List[Dict]:
        """Generate list of required clocks.

        For ADRV9009 this will contain:
        [device clock requirement SOS, sysref requirement SOS]

        Returns:
            list[dict]: List of dictionaries of solver variables, equations, and constants
        """
        if self.solver == "gekko":
            return self._gekko_get_required_clocks()
        self.config = {}
        self.config["lmfc_divisor_sysref"] = self._convert_input(
            [*range(1, 20)], name="lmfc_divisor_sysref"
        )

        self.config["input_clock_divider_x2"] = self._convert_input(
            self.input_clock_dividers_times2_available
        )
        self.config["ref_clk"] = self._add_intermediate(
            self.sample_clock / self.config["input_clock_divider_x2"]
        )
        self.config["sysref"] = self._add_intermediate(
            self.multiframe_clock
            / (self.config["lmfc_divisor_sysref"] * self.config["lmfc_divisor_sysref"])
        )

        self._add_equation(
            [
                self.device_clock_min <= self.config["ref_clk"],
                self.config["ref_clk"] <= self.device_clock_max,
            ]
        )

        return [self.config["ref_clk"], self.config["sysref"]]


class adrv9009_rx(adc, adrv9009_clock_common, adrv9009_core):
    """ADRV9009 Receive model."""

    quick_configuration_modes = {"jesd204b": quick_configuration_modes_rx}
    name = "ADRV9009_RX"
    converter_type = "adc"

    # JESD configurations
    K_available = [*np.arange(1, 32 + 1)]
    L_available = [1, 2, 4]
    M_available = [1, 2, 4]
    N_available = [12, 14, 16, 24]
    Np_available = [12, 16, 24]
    F_available = [
        1,
        2,
        3,
        4,
        6,
        8,
    ]
    CS_available = [0]
    CF_available = [0]
    S_available = [1, 2, 4]

    # Clock constraints
    bit_clock_min_available = {"jesd204b": 3.6864e9}
    bit_clock_max_available = {"jesd204b": 12.288e9}

    _decimation = 8
    decimation_available = [4, 5, 8, 10, 16, 20, 32, 40]
    """
    ADRV9009 Rx decimation stages.

    .. code-block:: none

            +-----------+
            +-----------+ Dec 5 (5) +---------+
            |           +-----------+         |
            |                                 |
            |   +----------+   +----------+   |  +------------+   +--------------+
        >---+---+ RHB3 (2) +---+ RHB2 (2) +---+--+ RHB1 (1,2) +---+ RFIR (1,2,4) +
                +----------+   +----------+      +------------+   +--------------+
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize ADRV9009-RX class.

        Objects will default to mode and sample_clock.

        Args:
            *args (Any): Pass through arguments
            **kwargs (Any): Pass through keyword arguments
        """
        # Set default mode
        self.set_quick_configuration_mode(str(16))
        # self.K = 32
        self.sample_clock = int(122.88e6)
        super().__init__(*args, **kwargs)
        # self._init_diagram()

    def get_required_clock_names(self) -> List[str]:
        """Get list of strings of names of requested clocks.

        This list of names is for the clocks defined by
        get_required_clocks

        Returns:
            List[str]: List of strings of clock names mapped by get_required_clocks
        """
        return ["adrv9009_rx_ref_clk", "adrv9009_rx_sysref"]


class adrv9009_tx(dac, adrv9009_clock_common, adrv9009_core):
    """ADRV9009 Transmit model."""

    quick_configuration_modes = {"jesd204b": quick_configuration_modes_tx}
    name = "ADRV9009_TX"
    converter_type = "dac"

    # JESD configurations
    K_available = [*np.arange(1, 32 + 1)]
    L_available = [1, 2, 4]
    M_available = [1, 2, 4]
    N_available = [12, 16]
    Np_available = [12, 16]
    F_available = [1, 2, 3, 4, 8]
    CS_available = [0]
    CF_available = [0]
    S_available = [1]

    # Clock constraints
    bit_clock_min_available = {"jesd204b": 2457.6e6}
    bit_clock_max_available = {"jesd204b": 12.288e9}

    _interpolation = 8
    interpolation_available = [1, 2, 4, 5, 8, 10, 16, 20, 32]
    """
    ADRV9009 Tx interpolation stages.

    .. code-block:: none

                                +------------+
            +--------------------+ Int 5  (5) +--------------------+
            |                    +------------+                    |
            |                                                      |
            |   +------------+   +------------+   +------------+   |   +--------------+
        <---+---+ THB3 (1,2) +---+ THB2 (1,2) +---+ THB1 (1,2) +---+---+ TFIR (1,2,4) +
                +------------+   +------------+   +------------+       +--------------+

    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize ADRV9009-TX class.

        Objects will default to mode and sample_clock.

        Args:
            *args (Any): Pass through arguments
            **kwargs (Any): Pass through keyword arguments
        """
        # Set default mode
        self.set_quick_configuration_mode(str(6))
        self.sample_clock = int(122.88e6)
        super().__init__(*args, **kwargs)
        # self._init_diagram()

    def get_required_clock_names(self) -> List[str]:
        """Get list of strings of names of requested clocks.

        This list of names is for the clocks defined by
        get_required_clocks

        Returns:
            List[str]: List of strings of clock names mapped by get_required_clocks
        """
        return ["adrv9009_tx_ref_clk", "adrv9009_tx_sysref"]


class adrv9009(adrv9009_core):
    """ADRV9009 combined transmit and receive model."""

    name = "ADRV9009"
    solver = "CPLEX"
    _nested = ["adc", "dac"]
    converter_type = "adc_dac"

    def __init__(
        self, model: Union[GEKKO, CpoModel] = None, solver: str = None
    ) -> None:
        """Initialize ADRV9009 clocking model for TX and RX.

        This is a common class used to handle TX and RX constraints
        together.

        Args:
            model (GEKKO,CpoModel): Solver model
            solver (str): Solver name (gekko or CPLEX)
        """
        if solver:
            self.solver = solver
        self.adc = adrv9009_rx(model, solver=self.solver)
        self.dac = adrv9009_tx(model, solver=self.solver)
        self.model = model

    def validate_config(self) -> None:
        """Validate device configurations including JESD and clocks of both ADC and DAC.

        This check only is for static configuration that does not include
        variables which are solved.
        """
        self.adc.validate_config()
        self.dac.validate_config()

    def _get_converters(self) -> List[Union[converter, converter]]:
        return [self.adc, self.dac]

    def get_required_clocks(self) -> List[Dict]:
        """Generate list of required clocks.

        For ADRV9009 this will contain:
        [device clock requirement SOS, sysref requirement SOS]

        Returns:
            list[dict]: List of dictionaries of solver variables, equations, and constants

        Raises:
            Exception: Invalid relation of rates between RX and TX
            AssertionError: Gekko called
        """
        # Validate sample rates feasible
        if self.dac.sample_clock / self.adc.sample_clock not in [
            1,
            2,
            4,
        ] or self.adc.sample_clock / self.dac.sample_clock not in [1, 2, 4]:
            raise Exception(
                "ADRV9009 RX and TX sample rates must be related by power of 2"
            )

        if self.solver == "gekko":
            raise AssertionError

        self.config = {}
        self.config["adc_lmfc_divisor_sysref"] = self._convert_input(
            self._lmfc_divisor_sysref_available, name="adc_lmfc_divisor_sysref"
        )
        self.config["dac_lmfc_divisor_sysref"] = self._convert_input(
            self._lmfc_divisor_sysref_available, name="dac_lmfc_divisor_sysref"
        )

        self.config["input_clock_divider_x2"] = self._convert_input(
            self.input_clock_dividers_times2_available
        )

        faster_clk = max([self.adc.sample_clock, self.dac.sample_clock])
        self.config["ref_clk"] = self._add_intermediate(
            faster_clk / self.config["input_clock_divider_x2"]
        )

        self.config["sysref_adc"] = self._add_intermediate(
            self.adc.multiframe_clock / self.config["adc_lmfc_divisor_sysref"]
        )
        self.config["sysref_dac"] = self._add_intermediate(
            self.dac.multiframe_clock / self.config["dac_lmfc_divisor_sysref"]
        )

        self._add_equation(
            [
                self.device_clock_min <= self.config["ref_clk"],
                self.config["ref_clk"] <= self.device_clock_max,
            ]
        )

        return [
            self.config["ref_clk"],
            self.config["sysref_adc"],
            self.config["sysref_dac"],
        ]



# flake8: noqa
import pprint

import pytest

import adijif


def test_adf4371_datasheet_example():
    pll = adijif.adf4371()
    pll._MOD2 = 1536
    pll.rf_div = 2
    pll.mode = "fractional"

    ref_in = int(122.88e6)
    output_clocks = int(2112.8e6)

    pll.set_requested_clocks(ref_in, output_clocks)

    pll.solve()

    o = pll.get_config()

    pprint.pprint(o)

    D = o["d"]
    R = o["r"]
    T = o["t"]
    INT = o["int"]
    FRAC1 = o["frac1"]
    FRAC2 = o["frac2"]
    MOD2 = o["MOD2"]
    rf_div = o["rf_div"]

    MOD1 = 2**25

    F_PFD = ref_in * (1 + D) / (R * (1 + T))
    vco = (INT + (FRAC1 + FRAC2 / MOD2) / MOD1) * F_PFD

    assert int(vco) == int(2112.8e6 * rf_div)
    assert int(vco / rf_div) == int(output_clocks)


def test_adf4371_ad9081_sys_example():
    vcxo = 100e6
    cddc = 6
    fddc = 4

    sys = adijif.system("ad9081", "hmc7044", "xilinx", vcxo, solver="CPLEX")
    sys.fpga.setup_by_dev_kit_name("zcu102")
    sys.fpga.ref_clock_constraint = "Unconstrained"
    sys.fpga.sys_clk_select = "XCVR_QPLL0"  # Use faster QPLL
    sys.fpga.out_clk_select = (
        "XCVR_PROGDIV_CLK"  # force reference to be core clock rate
    )
    sys.converter.adc.sample_clock = 2900000000 / (cddc * fddc)
    sys.converter.dac.sample_clock = 5800000000 / (cddc * fddc)
    # sys.converter.adc.decimation = 8 * 6
    # sys.converter.dac.interpolation = 4 * 12
    sys.converter.adc.datapath.cddc_decimations = [cddc] * 4
    sys.converter.adc.datapath.fddc_decimations = [fddc] * 8
    sys.converter.adc.datapath.fddc_enabled = [True] * 8
    sys.converter.dac.datapath.cduc_interpolation = cddc
    sys.converter.dac.datapath.fduc_interpolation = fddc
    sys.converter.dac.datapath.fduc_enabled = [True] * 8
    assert sys.converter.dac.interpolation == cddc * fddc

    sys.converter.clocking_option = "direct"
    sys.add_pll_inline("adf4371", sys.clock, sys.converter)

    mode_tx = "0"
    mode_rx = "1.0"

    sys.converter.dac.set_quick_configuration_mode(mode_tx, "jesd204c")
    sys.converter.adc.set_quick_configuration_mode(mode_rx, "jesd204c")

    assert sys.converter.adc.M == 8
    assert sys.converter.adc.F == 12
    assert sys.converter.adc.K == 64
    assert sys.converter.adc.Np == 12
    assert sys.converter.adc.CS == 0
    assert sys.converter.adc.L == 1
    assert sys.converter.adc.S == 1

    sys.converter.adc._check_clock_relations()
    sys.converter.dac._check_clock_relations()

    cfg = sys.solve()

    pprint.pprint(cfg)
    print(sys.converter.dac.converter_clock)

    assert float(cfg["clock_ext_pll_adf4371"]["rf_out_frequency"]) == float(
        sys.converter.dac.converter_clock
    ), f"{cfg['clock_ext_pll_adf4371']['rf_out_frequency']} != {sys.converter.dac.converter_clock}"


@pytest.mark.parametrize(
    "mode",
    [
        "integer",
        "fractional",
        ["integer", "fractional"],
    ],
)
@pytest.mark.parametrize(
    "int_prescaler",
    ["4/5", "8/9", ["4/5", "8/9"]],
)
def test_adf4371_vary_modes(mode, int_prescaler):
    pll = adijif.adf4371()
    pll.mode = mode
    pll._prescaler = int_prescaler

    # ref_in = int(122.88e6*2)
    # output_clocks = int(2112.8e6)
    ref_in = int(10e6)
    output_clocks = int(100e6)

    pll.set_requested_clocks(ref_in, output_clocks)

    pll.solve()

    o = pll.get_config()

    pprint.pprint(o)

    D = o["d"]
    R = o["r"]
    T = o["t"]
    INT = o["int"]
    FRAC1 = o["frac1"]
    FRAC2 = o["frac2"]
    MOD2 = o["MOD2"]
    rf_div = o["rf_div"]

    MOD1 = 2**25

    F_PFD = ref_in * (1 + D) / (R * (1 + T))
    vco = (INT + (FRAC1 + FRAC2 / MOD2) / MOD1) * F_PFD

    assert int(vco) == int(output_clocks * rf_div)


def test_adf4371_touch_all_properties():
    pll = adijif.adf4371()

    # read/write all
    pll.mode = pll.mode
    pll._prescaler = pll._prescaler
    pll.d = pll.d
    pll.r = pll.r
    pll.t = pll.t
    pll.rf_div = pll.rf_div

    # Manually set
    pll.mode = ["integer", "fractional"]
    pll._prescaler = ["4/5", "8/9", ["4/5", "8/9"]]
    pll.d = 0
    pll.r = 32
    pll.t = 0
    pll.rf_div = 2

    ref_in = int(122.88e6)
    output_clocks = int(2112.8e6)

    pll.set_requested_clocks(ref_in, output_clocks)

    pll.solve()

    o = pll.get_config()

    pprint.pprint(o)

    D = o["d"]
    R = o["r"]
    T = o["t"]
    INT = o["int"]
    FRAC1 = o["frac1"]
    FRAC2 = o["frac2"]
    MOD2 = o["MOD2"]
    rf_div = o["rf_div"]

    MOD1 = 2**25

    F_PFD = ref_in * (1 + D) / (R * (1 + T))
    vco = (INT + (FRAC1 + FRAC2 / MOD2) / MOD1) * F_PFD

    assert rf_div == 2
    assert int(vco) == int(2112.8e6 * rf_div)
    assert int(vco / rf_div) == int(output_clocks)


def test_wrong_solver():
    msg = "Only CPLEX solver is implemented"
    with pytest.raises(Exception, match=msg):
        pll = adijif.adf4371(solver="gekko")
        pll._MOD2 = 1536
        pll.rf_div = 2
        pll.mode = "fractional"

        ref_in = int(122.88e6)
        output_clocks = int(2112.8e6)

        pll.set_requested_clocks(ref_in, output_clocks)

        pll.solve()


def test_clock_names_not_set():
    msg = "set_requested_clocks must be called before get_config"
    with pytest.raises(Exception, match=msg):
        pll = adijif.adf4371()
        pll._MOD2 = 1536
        pll.rf_div = 2
        pll.mode = "fractional"

        ref_in = int(122.88e6)
        output_clocks = int(2112.8e6)

        pll.set_requested_clocks(ref_in, output_clocks)

        pll.solve()

        pll._clk_names = None

        pll.get_config()



# flake8: noqa
# pytype: skip-file
import numpy as np

from adijif.clocks.clock import clock


class ad9523_1_bf(clock):
    """Brute force methods for calculating clocks

    These are currently meant for debug to compare against
    the solver solutions
    """

    def list_available_references(self, divider_set):
        """list_available_references: Based on config list possible
        references that can be generated based on VCO and output
        dividers
        """
        # Check input
        ref = {
            "m1": 3,
            "vco": 3000000000,
            "r2": 24,
            "required_output_divs": np.array([1.0]),
        }
        for key in ref:
            if key not in divider_set:
                raise Exception(
                    "Input must be of type dict with fields: " + str(ref.keys())
                )
        return [
            divider_set["vco"] / divider_set["m1"] / div for div in self.d_available
        ]

    def find_dividers(self, vcxo, required_output_rates, find=3):
        if self.use_vcxo_double:
            vcxo *= 2

        # even =  np.arange(2, 4096, 2, dtype=int)
        # odivs = np.append([1, 3, 5], even)

        mod = np.gcd.reduce(np.array(required_output_rates, dtype=int))
        configs = []

        for n2 in self.n2_available:
            for r2 in self.r2_available:
                pfb = vcxo / r2
                if pfb > self.pfd_max:
                    continue
                vco = pfb * n2
                if vco > self.vco_min and vco < self.vco_max:
                    for m1 in self.m1_available:
                        # print("vco",vco,mod,m1)
                        if (vco / m1) % mod == 0:
                            # See if we can use only m1 and not both m1+m2
                            rods = (vco / m1) / required_output_rates
                            if np.all(np.isin(rods, self.d_available)):
                                configs.append(
                                    {
                                        "m1": m1,
                                        "n2": n2,
                                        "vco": vco,
                                        "r2": r2,
                                        "required_output_divs": rods,
                                    }
                                )
                            else:
                                # Try to use m2 as well to meet required out clocks
                                f1 = np.isin(rods, self.d_available)
                                for m2 in self.m1_available:
                                    rods2 = (vco / m2) / required_output_rates
                                    f2 = np.isin(rods2, self.d_available)
                                    if np.logical_or(f1, f2).all():
                                        configs.append(
                                            {
                                                "m1": m1,
                                                "m2": m2,
                                                "n2": n2,
                                                "vco": vco,
                                                "r2": r2,
                                                "required_output_divs": rods[
                                                    f1
                                                ].tolist(),
                                                "required_output_divs2": rods2[
                                                    f2
                                                ].tolist(),
                                            }
                                        )

        return configs



"""ADF4382 Microwave Wideband Synthesizer with Integrated VCO model."""

from typing import Dict, List, Union

from docplex.cp.modeler import if_then
from docplex.cp.solution import CpoSolveResult  # type: ignore

from adijif.plls.pll import pll
from adijif.solvers import CpoExpr, GK_Intermediate, integer_var


def to_int(value: Union[int, float, List[int], List[float]]) -> Union[int, List[int]]:
    """Convert value to int or list of ints."""
    if isinstance(value, (int, float)):
        return int(value)
    elif isinstance(value, list):
        return [int(v) for v in value]
    else:
        raise TypeError(f"Unsupported type: {type(value)}")


class adf4382(pll):
    """ADF4382 PLL model.

    This model does not support fractional mode

    https://www.analog.com/media/en/technical-documentation/data-sheets/adf4382.pdf
    """

    name = "adf4382"

    input_freq_min = int(10e6)
    input_freq_max = int(4.5e9)

    # pfd_freq_max_frac = int(160e6)
    """Input reference doubler range"""
    freq_doubler_input_min = int(10e6)
    freq_doubler_input_max = int(2000e6)

    """PFD frequency ranges"""
    # Integer
    pfd_freq_min_int_n_wide = int(5.4e6)
    pfd_freq_max_int_n_wide = int(625e6)
    # n full [*range(4, 4095+1)] minus 15, 28, 31
    _n_int_wide = (
        [*range(4, 15)] + [*range(16, 28)] + [*range(29, 31)] + [*range(32, 4095 + 1)]
    )
    pfd_freq_min_int_n_narrow = int(5.4e6)
    pfd_freq_max_int_n_narrow = int(540e6)
    _n_int_narrow = [15, 28, 31]
    # Fractional
    # n full [*range(19, 4095+1)]
    pfd_freq_min_frac_modes_0_4 = int(5.4e6)
    pfd_freq_max_frac_modes_0_4 = int(250e6)
    _n_frac_modes_0_4 = {
        "mode_0": to_int([*range(10, 4095 + 1)]),
        "mode_4": to_int([*range(23, 4095 + 1)]),
    }
    pfd_freq_min_frac_modes_5 = int(5.4e6)
    pfd_freq_max_frac_modes_5 = int(220e6)
    _n_frac_modes_5 = to_int([*range(27, 4095 + 1)])

    vco_freq_min = int(11e9)
    vco_freq_max = int(22e9)

    _d = to_int([1, 2])
    d_available = [1, 2]

    @property
    def d(self) -> Union[int, List[int]]:
        """REF-in doubler.

        Valid values are 1,2

        Returns:
            int: Current allowable setting
        """
        return self._d

    @d.setter
    def d(self, value: Union[int, List[int]]) -> None:
        """REF-in double.

        Valid values are 1,2

        Args:
            value (int, list[int]): Current allowable setting

        """
        self._check_in_range(value, self.d_available, "d")
        self._d = value

    _r = [*range(1, 63 + 1)]
    r_available = [*range(1, 63 + 1)]

    @property
    def r(self) -> Union[int, List[int]]:
        """Reference divider.

        Valid values are 1->(2^6-1)

        Returns:
            int: Current allowable setting
        """
        return self._r

    @r.setter
    def r(self, value: Union[int, List[int]]) -> None:
        """Reference divider.

        Valid values are 1->63

        Args:
            value (int, list[int]): Current allowable setting

        """
        self._check_in_range(value, self.r_available, "r")
        self._r = value

    _o = [1, 2, 4]
    o_available = [1, 2, 4]

    @property
    def o(self) -> Union[int, List[int]]:
        """Output RF divider.

        Valid dividers are 1,2,4

        Returns:
            int: Current allowable dividers
        """
        return self._o

    @o.setter
    def o(self, value: Union[int, List[int]]) -> None:
        """Output RF divider.

        Valid dividers are 1,2,4

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.o_available, "o")
        self._o = value

    _n = [*range(4, 2**12)]
    n_available = [*range(4, 2**12)]

    @property
    def n(self) -> Union[int, List[int]]:
        """Feedback divider.

        Valid dividers are 1->4096

        Returns:
            int: Current allowable dividers
        """
        return self._n

    @n.setter
    def n(self, value: Union[int, List[int]]) -> None:
        """Feedback divider.

        Valid dividers are 1->4096

        Args:
            value (int, list[int]): Allowable values for divider

        """
        self._check_in_range(value, self.n_available, "n")
        self._n = value

    _mode = ["integer"]  # Dont use fractional mode by default
    mode_available = ["integer", "fractional"]

    @property
    def mode(self) -> Union[str, List[str]]:
        """Set operational mode.

        Options are: fractional, integer or [fractional, integer]

        Returns:
            str: Current allowable modes
        """
        return self._mode

    @mode.setter
    def mode(self, value: Union[str, List[str]]) -> None:
        """Set operational mode.

        Options are: fractional, integer or [fractional, integer]

        Args:
            value (str, list[str]): Allowable values for mode

        """
        self._check_in_range(value, self.mode_available, "mode")
        self._mode = value

    # These are too large for user to set
    _frac1_min_max = [0, 2**25 - 1]
    _frac2_min_max = [0, 2**24 - 1]
    _MOD1 = 33554432  # 2^25
    _MOD2_PHASE_SYNC_min_max = [1, 2**17 - 1]
    _MOD2_NO_PHASE_SYNC_min_max = [1, 2**24 - 1]

    _phase_sync = True

    @property
    def require_phase_sync(self) -> bool:
        """Determine if phase sync is required.

        Returns:
            bool: True if phase sync is required
        """
        return self._phase_sync

    @require_phase_sync.setter
    def require_phase_sync(self, value: bool) -> None:
        """Determine if phase sync is required.

        Args:
            value (bool): True if phase sync is required
        """
        self._check_in_range(value, [True, False], "require_phase_sync")
        self._phase_sync = value

    _EFM3_MODE = [0, 4, 5]
    EFM3_MODE_available = [0, 4, 5]

    @property
    def EFM3_MODE(self) -> Union[int, List[int]]:
        """Set EFM3 optimization mode.

        Options are: 0, 4, 5

        Returns:
            int: Current allowable modes
        """
        return self._EFM3_MODE

    @EFM3_MODE.setter
    def EFM3_MODE(self, value: Union[int, List[int]]) -> None:
        """Set EFM3 optimization mode.

        Options are: 0, 4, 5

        Args:
            value (int, list[int]): Allowable values for mode

        """
        self._check_in_range(value, self.EFM3_MODE_available, "EFM3_MODE")
        self._EFM3_MODE = value

    def get_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect internal clock chip configuration and output clock definitions
        leading to connected devices (converters, FPGAs)

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration

        Raises:
            Exception: If solver is not called first
        """
        if not self._clk_names:
            raise Exception("set_requested_clocks must be called before get_config")

        if solution:
            self.solution = solution

        config: Dict = {
            "d": self._get_val(self.config["d"]),
            # "n": self.solution.get_kpis()["n"],
            "o": self._get_val(self.config["o"]),
            "r": self._get_val(self.config["r"]),
        }
        if isinstance(self._n, list):
            config["n"] = self.solution.get_kpis()["n"]
        else:
            config["n"] = self._n

        if self._mode == "integer" or self._mode == ["integer"]:
            mode = "integer"
        elif self._mode == "fractional" or self._mode == ["fractional"]:
            mode = "fractional"
        else:
            if self._get_val(self.config["frac_0_int_1"]) == 1:
                mode = "integer"
            else:
                mode = "fractional"

        if mode == "integer":
            config["mode"] = "integer"
        else:
            config["mode"] = "fractional"
            config["n_frac1w"] = self._get_val(self.config["n_frac1w"])
            config["n_frac2w"] = self._get_val(self.config["n_frac2w"])
            config["MOD1"] = self._MOD1
            config["MOD2"] = self._get_val(self.config["MOD2"])
            config["n_int"] = self._get_val(self.config["n_int"])
            config["EFM3_MODE"] = self._get_val(self.config["EFM3_MODE"])

        vco = self.solution.get_kpis()["vco"]
        config["rf_out_frequency"] = vco / config["o"]

        return config

    def _setup_solver_constraints(
        self, input_ref: Union[int, float, CpoExpr, GK_Intermediate]
    ) -> None:
        """Apply constraints to solver model.

        Args:
            input_ref (int, float, CpoExpr, GK_Intermediate): Input reference
                frequency in hertz. Can also be range or arb_source type.

        Raises:
            NotImplementedError: If solver is not CPLEX
        """
        self.config = {}

        if self.solver != "CPLEX":
            raise NotImplementedError("Only CPLEX solver is supported")

        # Handle range type (returns dict with "range" key)
        # arb_source and direct expressions are already supported
        expr_types = tuple(filter(None, [int, float, CpoExpr, GK_Intermediate]))
        if not isinstance(input_ref, expr_types):
            input_ref_result = input_ref(self.model)  # type: ignore
            if isinstance(input_ref_result, dict):
                # range type
                self.config["input_ref_set"] = input_ref_result
                input_ref = self.config["input_ref_set"]["range"]
            else:
                # arb_source type (returns expression directly)
                input_ref = input_ref_result
        self.input_ref = input_ref

        # PFD
        self.config["d"] = self._convert_input(self.d, name="d")
        self.config["r"] = self._convert_input(self.r, name="r")
        self.config["o"] = self._convert_input(self.o, name="o")

        if self._mode == "integer":
            self.config["frac_0_int_1"] = 1
        elif self._mode == "fractional":
            self.config["frac_0_int_1"] = 0
        else:
            self.config["frac_0_int_1"] = self._convert_input(
                [0, 1], name="frac_0_int_1"
            )

        self.config["EFM3_MODE"] = self._convert_input(self.EFM3_MODE, name="EFM3_MODE")
        if isinstance(self.EFM3_MODE, list):
            self.model.add_kpi(self.config["EFM3_MODE"], "EFM3_MODE")

        # N which supports fractional modes
        if "fractional" in self._mode:
            self.config["n_int"] = self._convert_input(self._n, name="n_int")
            self.config["n_frac1w"] = integer_var(
                min=self._frac1_min_max[0],
                max=self._frac1_min_max[1],
                name="n_frac1w",
            )
            self.config["n_frac2w"] = integer_var(
                min=self._frac2_min_max[0],
                max=self._frac2_min_max[1],
                name="n_frac2w",
            )
            _MOD2_min_max = (
                self._MOD2_PHASE_SYNC_min_max
                if self._phase_sync
                else self._MOD2_NO_PHASE_SYNC_min_max
            )
            self.config["MOD2"] = integer_var(
                min=_MOD2_min_max[0],
                max=_MOD2_min_max[1],
                name="MOD2",
            )
            self.config["n_frac"] = self._add_intermediate(
                (
                    self.config["n_frac1w"]
                    + self.config["n_frac2w"] / self.config["MOD2"]
                )
                / self._MOD1
            )
            # Constraints
            self._add_equation(
                if_then(
                    self.config["frac_0_int_1"] == 1,
                    self.config["n_frac"] == 0,
                )
            )
            self._add_equation(
                if_then(
                    self.config["frac_0_int_1"] == 0,
                    100000 * self.config["n_frac"] < 100000,
                )
            )
            self._add_equation(
                if_then(
                    self.config["frac_0_int_1"] == 0,
                    self.config["n_frac1w"] + self.config["n_frac2w"] > 0,
                )
            )
            self._add_equation(
                if_then(
                    self.config["frac_0_int_1"] == 0,
                    self.config["MOD2"] > self.config["n_frac2w"],
                )
            )

            self.config["n"] = self._add_intermediate(
                self.config["n_int"] + self.config["n_frac"]
            )
        else:
            self.config["n"] = self._convert_input(self._n, name="n_int")

        if isinstance(self._mode, list):
            # self.model.maximize(self.config["frac_0_int_1"])
            self._add_objective(self.config["frac_0_int_1"])

        # Add PFD frequency dependent on N
        if isinstance(self._n, list) and len(self._n) > 1:
            self.model.add_kpi(
                self.config["n"],
                "n",
            )

        # Add EFM3 mode constraints on N
        self._add_equation(
            [
                if_then(
                    self.config["EFM3_MODE"] == 0,
                    self.config["n"] >= min(self._n_frac_modes_0_4["mode_0"]),
                ),
                if_then(
                    self.config["EFM3_MODE"] == 4,
                    self.config["n"] >= min(self._n_frac_modes_0_4["mode_4"]),
                ),
                if_then(
                    self.config["EFM3_MODE"] == 5,
                    self.config["n"] >= min(self._n_frac_modes_5),
                ),
            ]
        )

        # Clocking rates
        self.config["f_pfd"] = self._add_intermediate(
            input_ref * self.config["d"] / self.config["r"]
        )

        self.config["vco"] = self._add_intermediate(
            self.config["f_pfd"] * self.config["n"] * self.config["o"]
        )
        self.model.add_kpi(
            self.config["f_pfd"] * self.config["n"] * self.config["o"],
            "vco",
        )

        # Add PFD frequency constraints for integer mode
        self._add_equation(
            [
                if_then(
                    self.config["frac_0_int_1"] == 1,
                    if_then(
                        self.config["n"] == 15,
                        self.config["f_pfd"] <= self.pfd_freq_max_int_n_narrow,
                    ),
                ),
                if_then(
                    self.config["frac_0_int_1"] == 1,
                    if_then(
                        self.config["n"] == 28,
                        self.config["f_pfd"] <= self.pfd_freq_max_int_n_narrow,
                    ),
                ),
                if_then(
                    self.config["frac_0_int_1"] == 1,
                    if_then(
                        self.config["n"] == 31,
                        self.config["f_pfd"] <= self.pfd_freq_max_int_n_narrow,
                    ),
                ),
                # Wide is a looser upper bound but applies to all cases
                if_then(
                    self.config["frac_0_int_1"] == 1,
                    self.config["f_pfd"] <= self.pfd_freq_max_int_n_wide,
                ),
            ]
        )

        # Add PFD frequency constraints for fractional mode
        self._add_equation(
            [
                if_then(
                    self.config["frac_0_int_1"] == 0,
                    if_then(
                        self.config["EFM3_MODE"] == 0,
                        self.config["f_pfd"] <= self.pfd_freq_max_frac_modes_0_4,
                    ),
                ),
                if_then(
                    self.config["frac_0_int_1"] == 0,
                    if_then(
                        self.config["EFM3_MODE"] == 4,
                        self.config["f_pfd"] <= self.pfd_freq_max_frac_modes_0_4,
                    ),
                ),
                if_then(
                    self.config["frac_0_int_1"] == 0,
                    if_then(
                        self.config["EFM3_MODE"] == 5,
                        self.config["f_pfd"] <= self.pfd_freq_max_frac_modes_5,
                    ),
                ),
            ]
        )

        # Global min for PFD
        self._add_equation(
            [
                self.config["f_pfd"] >= self.pfd_freq_min_int_n_wide,
            ]
        )

        # Add remaining constraints
        self._add_equation(
            [
                input_ref <= self.input_freq_max,
                input_ref >= self.input_freq_min,
                input_ref * self.config["d"] <= self.freq_doubler_input_max,
                input_ref * self.config["d"] >= self.freq_doubler_input_min,
                self.config["vco"] <= self.vco_freq_max,
                self.config["vco"] >= self.vco_freq_min,
            ]
        )

        # Minimize feedback divider to reduce jitter
        self._add_objective(1 / self.config["n"])
        # self.model.minimize(self.config['n'])

    def _setup(self, input_ref: int) -> None:
        if isinstance(input_ref, (float, int)):
            assert (
                self.input_freq_max >= input_ref >= self.input_freq_min
            ), "Input frequency out of range"

        # Setup clock chip internal constraints
        self._setup_solver_constraints(input_ref)

    def _get_clock_constraint(
        self, clk_name: str
    ) -> Union[int, float, CpoExpr, GK_Intermediate]:
        """Get abstract clock output.

        Args:
            clk_name (str):  Reference clock name

        Returns:
            (int or float or CpoExpr or GK_Intermediate): Abstract
                or concrete clock reference
        """
        self._clk_names = ["clk_name"]

        assert "o" in self.config, "_setup must be called first to set PLL internals"

        return self.config["vco"] / self.config["o"]

    def set_requested_clocks(
        self, ref_in: Union[int, float, CpoExpr, GK_Intermediate], out_freq: int
    ) -> None:
        """Define necessary clocks to be generated in model.

        Args:
            ref_in (int, float, CpoExpr, GK_Intermediate): Reference frequency in hertz
            out_freq (int): list of required clocks to be output

        """
        self._setup(ref_in)
        self._clk_names = ["rf_out"]

        self._add_equation([self.config["o"] * out_freq == self.config["vco"]])



"""Diagram drawing for System level."""

from typing import Dict

from adijif.draw import Layout, Node  # type: ignore # isort: skip  # noqa: I202


class system_draw:
    """Drawing features for system level models."""

    def _init_diagram(self) -> None:
        """Initialize diagram for system."""
        self.ic_diagram_node = None

        self.ic_diagram_node = Node("System")

    def draw(self, config: Dict) -> str:
        """Draw clocking model for system.

        Args:
            config(Dict): System solution configuration

        Returns:
            str: Drawn diagram
        """
        lo = Layout("System Diagram")
        if hasattr(self, "use_d2_cli"):
            lo.use_d2_cli = self.use_d2_cli
        self._init_diagram()

        self.ic_diagram_node.add_child(lo)

        # Clocking
        assert self.clock is not None
        assert not isinstance(self.clock, list), "Only one clocking supported"

        self.clock.draw(lo)

        # Converter
        assert self.converter is not None
        assert not isinstance(self.converter, list), "Only one converter supported"

        cnv_clocking = config["clock"]["output_clocks"].copy()
        for clk in cnv_clocking:
            rate = cnv_clocking[clk]["rate"]
            cnv_clocking[clk] = rate

        self.converter.draw(cnv_clocking, lo, self.clock.ic_diagram_node)

        # FPGA
        assert self.fpga is not None
        assert not isinstance(self.fpga, list), "Only one FPGA supported"

        print(config)

        fpga_clocking = {
            "clocks": cnv_clocking,
            "fpga": config[f"fpga_{self.converter.name}"],
        }
        if not isinstance(self.converter, list):
            cnvers = [self.converter]
        self.fpga.draw(fpga_clocking, lo, cnvers)

        # Draw the diagram
        return lo.draw()



# flake8: noqa
from adijif.fpgas.fpga import fpga


class xilinx_bf(fpga):
    """Brute force methods for calculating clocks

    These are currently meant for debug to compare against
    the solver solutions
    """

    def determine_cpll(self, bit_clock, fpga_ref_clock):
        """
        Parameters:
            bit_clock:
                Equivalent to lane rate in bits/second
            fpga_ref_clock:
                System reference clock
        """
        assert isinstance(bit_clock, int), "bit_clock must be an int"
        assert isinstance(fpga_ref_clock, int), "fpga_ref_clock must be an int"

        # VCO = ( REF_CLK * N1 * N2 ) / M
        # bit_clock = ( VCO * 2 ) / D

        for m in [1, 2]:
            for d in [1, 2, 4, 8]:
                for n1 in [5, 4]:
                    for n2 in [5, 4, 3, 2, 1]:
                        vco = fpga_ref_clock * n1 * n2 / m
                        # print("VCO", self.vco_min/1e9, vco/1e9, self.vco_max/1e9)
                        if vco > self.vco_max or vco < self.vco_min:
                            continue
                        # print("VCO", vco)
                        # fpga_lane_rate = vco * 2 / d
                        # print("lane rate", fpga_lane_rate)

                        # VCO == 5,10,20,40 GHz

                        # print(fpga_ref_clock / m / d, bit_clock / (2 * n1 * n2))
                        if fpga_ref_clock / m / d == bit_clock / (2 * n1 * n2):
                            return {
                                "vco": vco,
                                "d": d,
                                "m": m,
                                "n1": n1,
                                "n2": n2,
                                "type": "CPLL",
                            }

        raise Exception("No valid CPLL configuration found")

    def determine_qpll(self, bit_clock, fpga_ref_clock):
        """
        Parameters:
            bit_clock:
                Equivalent to lane rate in bits/second
            fpga_ref_clock:
                System reference clock
        """

        if self.ref_clock_max < fpga_ref_clock or fpga_ref_clock < self.ref_clock_min:
            raise Exception("fpga_ref_clock not within range")

        for m in [1, 2, 3, 4]:
            for d in [1, 2, 4, 8, 16]:
                for n in self.N:
                    vco = fpga_ref_clock * n / m
                    if self.vco1_min <= vco <= self.vco1_max:
                        band = 1
                    elif self.vco0_min <= vco <= self.vco0_max:
                        band = 0
                    else:
                        continue

                    if fpga_ref_clock / m / d == bit_clock / n:
                        return {
                            "vco": vco,
                            "band": band,
                            "d": d,
                            "m": m,
                            "n": n,
                            "qty4_full_rate": 0,
                            "type": "QPLL",
                        }

                    if self.transciever_type != "GTY4":
                        continue

                    if fpga_ref_clock / m / d == bit_clock / 2 / n:
                        return {
                            "vco": vco,
                            "band": band,
                            "d": d,
                            "m": m,
                            "n": n,
                            "qty4_full_rate": 1,
                            "type": "QPLL",
                        }

        raise Exception("No valid QPLL configuration found")



# flake8: noqa
import pprint

import pytest

import adijif


def test_jesd_unknown_dev():
    name = "zcu103"
    with pytest.raises(Exception, match=f"No boardname found in library for {name}"):
        f = adijif.xilinx()
        f.setup_by_dev_kit_name(name)


def test_jesd_wrong_order():
    msg = (
        "get_required_clocks must be run to generated"
        + " dependent clocks before names are available"
    )
    with pytest.raises(Exception, match=msg):
        f = adijif.xilinx()
        f.ref_clock_min = 60000000
        f.ref_clock_max = 670000000
        cn = f.get_required_clock_names()


def test_ref_clock_not_set():
    msg = "ref_clock_min or ref_clock_max not set"
    with pytest.raises(Exception, match=msg):
        f = adijif.xilinx()
        cn = f.get_required_clocks([], [])


def test_xilinx_transceiver_generation():
    """Test xilinx transceiver generation method."""
    f = adijif.xilinx()
    f.transceiver_type = "GTXE2"

    gen = f.trx_gen()
    assert gen == 2

    f.transceiver_type = "GTHE3"
    gen = f.trx_gen()
    assert gen == 3


def test_xilinx_speed_grades():
    """Test xilinx speed grade configuration."""
    f = adijif.xilinx()

    # Test valid speed grade
    f.speed_grade = -3
    assert f.speed_grade in f.available_speed_grades

    # Test another valid speed grade
    f.speed_grade = -1
    assert f.speed_grade in f.available_speed_grades


def test_xilinx_force_cpll():
    """Test xilinx CPLL forcing attribute."""
    sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6)
    sys.fpga.setup_by_dev_kit_name("zc706")

    # Test that attribute can be set
    sys.fpga.force_cpll = True
    assert sys.fpga.force_cpll is True


def test_xilinx_force_separate_device_clock():
    """Test xilinx with forced separate device clock attribute."""
    sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6)
    sys.fpga.setup_by_dev_kit_name("zc706")

    # Test that attribute can be set
    sys.fpga.force_separate_device_clock = True
    assert sys.fpga.force_separate_device_clock is True


def test_xilinx_ref_clock_constraint():
    """Test xilinx reference clock constraint options."""
    sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6)
    sys.fpga.setup_by_dev_kit_name("zc706")

    # Test that constraint can be set
    sys.fpga._ref_clock_constraint = "CORE_CLOCK_DIV2"
    assert sys.fpga._ref_clock_constraint == "CORE_CLOCK_DIV2"


def test_xilinx_favor_cpll_over_qpll():
    """Test xilinx with favor CPLL over QPLL attribute."""
    sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6)
    sys.fpga.setup_by_dev_kit_name("zc706")

    # Test that attribute can be set
    sys.fpga.favor_cpll_over_qpll = True
    assert sys.fpga.favor_cpll_over_qpll is True


def test_xilinx_minimize_fpga_ref_clock():
    """Test xilinx with minimize FPGA ref clock attribute."""
    sys = adijif.system("ad9680", "hmc7044", "xilinx", 125e6)
    sys.fpga.setup_by_dev_kit_name("zc706")

    # Test that attribute can be set
    sys.fpga.minimize_fpga_ref_clock = True
    assert sys.fpga.minimize_fpga_ref_clock is True


def test_xilinx_package_and_family():
    """Test xilinx package and family settings."""
    f = adijif.xilinx()

    # Test valid package
    f.fpga_package = "FF"
    assert f.fpga_package in f.available_fpga_packages

    # Test valid family
    f.fpga_family = "Kintex"
    assert f.fpga_family in f.available_fpga_families



# flake8: noqa
import pprint

import pytest

try:
    from gekko import GEKKO  # type: ignore
except ImportError:
    GEKKO = None

import adijif


def skip_solver(solver):
    if solver.lower() == "gekko" and GEKKO is None:
        pytest.skip("Gekko not available")


@pytest.mark.parametrize("solver", ["gekko", "CPLEX"])
def test_ad9545_validate_fail(solver):
    msg = r"Solution Not Found"

    skip_solver(solver)

    with pytest.raises(Exception, match=msg):
        clk = adijif.ad9545(solver=solver)

        clk.avoid_min_max_PLL_rates = True
        clk.minimize_input_dividers = True

        input_refs = [(0, 42345235)]
        output_clocks = [(0, 3525235234123)]

        input_refs = list(map(lambda x: (int(x[0]), int(x[1])), input_refs))
        output_clocks = list(map(lambda x: (int(x[0]), int(x[1])), output_clocks))

        clk.set_requested_clocks(input_refs, output_clocks)

        clk.solve()


@pytest.mark.parametrize("solver", ["gekko", "CPLEX"])
@pytest.mark.parametrize("out_freq", [30720000, 25e6])
def test_ad9545_validate_pass(solver, out_freq):
    skip_solver(solver)
    clk = adijif.ad9545(solver=solver)

    clk.avoid_min_max_PLL_rates = True
    clk.minimize_input_dividers = True

    input_refs = [(0, 1), (1, 10e6)]
    output_clocks = [(0, int(out_freq))]

    input_refs = list(map(lambda x: (int(x[0]), int(x[1])), input_refs))
    output_clocks = list(map(lambda x: (int(x[0]), int(x[1])), output_clocks))

    clk.set_requested_clocks(input_refs, output_clocks)

    clk.solve()
    sol = clk.get_config()

    PLLs_used = [False, False]
    for out_clock in output_clocks:
        if out_clock[0] > 5:
            PLLs_used[1] = True
        else:
            PLLs_used[0] = True

    for in_ref in input_refs:
        for pll_number in range(0, 2):
            if PLLs_used[pll_number]:
                pll_rate = sol["PLL" + str(pll_number)]["rate_hz"]
                n_name = "n" + str(pll_number) + "_profile_" + str(in_ref[0])
                assert (
                    pll_rate
                    == (in_ref[1] // sol["r" + str(in_ref[0])])
                    * sol["PLL" + str(pll_number)][n_name]
                )

    for out_clock in output_clocks:
        if out_clock[0] > 5:
            pll_rate = sol["PLL1"]["rate_hz"]
        else:
            pll_rate = sol["PLL0"]["rate_hz"]

        assert out_clock[1] == pll_rate / sol["q" + str(out_clock[0])]


def test_ad9545_fail_no_solver():
    with pytest.raises(Exception, match=r"Unknown solver NAN"):
        clk = adijif.ad9545(solver="NAN")

        input_refs = [(0, 1), (1, 10e6)]
        output_clocks = [(0, 30720000)]

        input_refs = list(map(lambda x: (int(x[0]), int(x[1])), input_refs))
        output_clocks = list(map(lambda x: (int(x[0]), int(x[1])), output_clocks))

        clk.set_requested_clocks(input_refs, output_clocks)

        clk.solve()


@pytest.mark.parametrize("solver", ["gekko", "CPLEX"])
def test_ad9523_1_daq2_validate(solver):
    skip_solver(solver)
    vcxo = 125000000
    n2 = 24

    clk = adijif.ad9523_1(solver=solver)

    # Check config valid
    clk.n2 = n2
    clk.use_vcxo_double = False

    output_clocks = [1e9, 500e6, 7.8125e6]
    clock_names = ["ADC", "FPGA", "SYSREF"]

    clk.set_requested_clocks(vcxo, output_clocks, clock_names)

    clk.solve()

    o = clk.get_config()

    assert sorted(o["out_dividers"]) == [1, 2, 128]
    assert o["m1"] == 3
    assert o["m1"] in clk.m1_available
    assert o["n2"] == n2
    assert o["n2"] in clk.n2_available
    assert o["r2"] == 1
    assert o["r2"] in clk.r2_available

    assert o["output_clocks"]["ADC"] == {"divider": 1, "rate": 1000000000.0}
    assert o["output_clocks"]["FPGA"] == {"divider": 2, "rate": 500000000.0}
    assert o["output_clocks"]["SYSREF"] == {"divider": 128, "rate": 7812500.0}


@pytest.mark.parametrize("solver", ["gekko", "CPLEX"])
def test_ad9523_1_daq2_validate_fail(solver):
    skip_solver(solver)
    msg = r"Solution Not Found"

    with pytest.raises(Exception, match=msg):
        vcxo = 125000000
        n2 = 12

        clk = adijif.ad9523_1(solver=solver)

        # Check config valid
        clk.n2 = n2
        # clk.r2 = 1
        clk.use_vcxo_double = False
        # clk.m = 3

        output_clocks = [1e9, 500e6, 7.8125e6]
        clock_names = ["ADC", "FPGA", "SYSREF"]

        clk.set_requested_clocks(vcxo, output_clocks, clock_names)

        o = clk.solve()

        print(o)

        assert sorted(o["out_dividers"]) == [
            2,
            4,
            256,
        ]  # This seems weird but its different per CPLEX version
        assert o["n2"] == n2


@pytest.mark.parametrize("solver", ["gekko", "CPLEX"])
def test_ad9523_1_daq2_variable_vcxo_validate(solver):
    skip_solver(solver)
    vcxo = adijif.types.range(100000000, 126000000, 1000000, "vcxo")
    n2 = 24

    clk = adijif.ad9523_1(solver=solver)

    # Check config valid
    clk.n2 = n2
    clk.use_vcxo_double = False

    output_clocks = [1e9, 500e6, 7.8125e6]
    clock_names = ["ADC", "FPGA", "SYSREF"]

    clk.set_requested_clocks(vcxo, output_clocks, clock_names)

    clk.solve()

    o = clk.get_config()

    print(o)

    assert sorted(o["out_dividers"]) == [1, 2, 128]
    assert o["m1"] == 3
    assert o["m1"] in clk.m1_available
    assert o["n2"] == n2
    assert o["n2"] in clk.n2_available
    assert o["r2"] == 1
    assert o["r2"] in clk.r2_available
    assert o["output_clocks"]["ADC"]["rate"] == 1e9
    assert o["vcxo"] == 125000000


def test_ad9523_1_fail_no_solver():
    with pytest.raises(Exception, match=r"Unknown solver NAN"):
        clk = adijif.ad9523_1(solver="NAN")
        output_clocks = [1e9, 500e6, 7.8125e6]
        clock_names = ["ADC", "FPGA", "SYSREF"]
        clk.set_requested_clocks(vcxo, output_clocks, clock_names)
        clk.solve()


def test_ad9523_1_fail_no_solver2():
    with pytest.raises(Exception, match=r"Unknown solver NAN2"):
        vcxo = 125000000
        clk = adijif.ad9523_1()
        clk.solver = "NAN2"
        output_clocks = [1e9, 500e6, 7.8125e6]
        clock_names = ["ADC", "FPGA", "SYSREF"]
        clk.set_requested_clocks(vcxo, output_clocks, clock_names)
        clk.solve()


def test_ad9523_1_fail_no_solver3():
    with pytest.raises(Exception, match=r"Unknown solver NAN3"):
        vcxo = 125000000
        clk = adijif.ad9523_1()
        output_clocks = [1e9, 500e6, 7.8125e6]
        clock_names = ["ADC", "FPGA", "SYSREF"]
        clk.set_requested_clocks(vcxo, output_clocks, clock_names)
        clk.solver = "NAN3"
        clk.solve()


def test_system_fail_no_solver3():
    with pytest.raises(Exception, match=r"Unknown solver NAN4"):
        vcxo = 125000000
        sys = adijif.system("ad9680", "hmc7044", "xilinx", vcxo, solver="NAN4")
        sys.solve()


def test_ltc6953_validate():
    ref_in = adijif.types.range(1000000000, 4500000000, 1000000, "ref_in")

    clk = adijif.ltc6953(solver="CPLEX")

    output_clocks = [1e9, 500e6, 7.8125e6]
    print(output_clocks)
    output_clocks = list(map(int, output_clocks))  # force to be ints
    clock_names = ["ADC", "FPGA", "SYSREF"]

    clk.set_requested_clocks(ref_in, output_clocks, clock_names)

    clk.solve()

    o = clk.get_config()

    assert sorted(o["out_dividers"]) == [2, 4, 256]
    assert o["input_ref"] == 2000000000


@pytest.mark.parametrize("solver", ["gekko", "CPLEX"])
def test_ad9528_validate(solver):
    skip_solver(solver)
    n2 = 10
    vcxo = 122.88e6

    clk = adijif.ad9528(solver=solver)

    clk.n2 = n2
    clk.use_vcxo_double = False

    output_clocks = [245.76e6, 245.76e6]
    output_clocks = list(map(int, output_clocks))
    clock_names = ["ADC", "FPGA"]

    clk.set_requested_clocks(vcxo, output_clocks, clock_names)

    clk.solve()
    o = clk.get_config()

    assert sorted(o["out_dividers"]) == [5, 5]
    assert o["m1"] == 3
    assert o["m1"] in clk.m1_available
    assert o["n2"] == n2
    assert o["n2"] in clk.n2_available
    assert o["output_clocks"]["ADC"]["rate"] == 245.76e6
    assert o["output_clocks"]["FPGA"]["rate"] == 245.76e6
    assert o["vcxo"] == vcxo
    assert o["vco"] == 3686400000.0


@pytest.mark.parametrize("solver", ["gekko", "CPLEX"])
def test_ad9528_sysref(solver):
    skip_solver(solver)
    n2 = 10
    vcxo = 122.88e6

    clk = adijif.ad9528(solver=solver)

    clk.n2 = n2
    clk.k = [*range(500, 600)]  # FIXME gekko fails to find a solution without this.
    clk.use_vcxo_double = False

    clk.sysref = 120e3

    output_clocks = [245.76e6, 245.76e6]
    output_clocks = list(map(int, output_clocks))
    clock_names = ["ADC", "FPGA"]

    clk.set_requested_clocks(vcxo, output_clocks, clock_names)

    clk.solve()
    o = clk.get_config()

    assert o["k"] == 512
    assert o["sysref"] == 120e3


# Tests for arb_source support
def test_ad9523_1_arb_source_vcxo():
    """Test AD9523-1 with arb_source vcxo."""
    vcxo = adijif.types.arb_source(
        "vcxo", a_min=100000000, a_max=150000000, b_min=1, b_max=1
    )
    n2 = 24

    clk = adijif.ad9523_1(solver="CPLEX")

    clk.n2 = n2
    clk.use_vcxo_double = False

    output_clocks = [1e9, 500e6, 7.8125e6]
    clock_names = ["ADC", "FPGA", "SYSREF"]

    clk.set_requested_clocks(vcxo, output_clocks, clock_names)

    clk.solve()

    o = clk.get_config()

    assert sorted(o["out_dividers"]) == [1, 2, 128]
    assert o["n2"] == n2
    assert o["vcxo"] == 125000000  # Should find optimal vcxo


def test_ad9528_arb_source_vcxo():
    """Test AD9528 with arb_source vcxo."""
    vcxo = adijif.types.arb_source(
        "vcxo", a_min=100000000, a_max=150000000, b_min=1, b_max=1
    )

    clk = adijif.ad9528(solver="CPLEX")

    clk.use_vcxo_double = False

    output_clocks = [245.76e6, 245.76e6]
    output_clocks = list(map(int, output_clocks))
    clock_names = ["ADC", "FPGA"]

    clk.set_requested_clocks(vcxo, output_clocks, clock_names)

    clk.solve()
    o = clk.get_config()

    assert isinstance(o["vcxo"], (int, float))
    assert 100000000 <= o["vcxo"] <= 150000000


def test_hmc7044_arb_source_vcxo():
    """Test HMC7044 with arb_source vcxo."""
    vcxo = adijif.types.arb_source(
        "vcxo", a_min=100000000, a_max=150000000, b_min=1, b_max=1
    )

    clk = adijif.hmc7044(solver="CPLEX")

    clk.use_vcxo_double = True

    output_clocks = [245.76e6, 245.76e6]
    output_clocks = list(map(int, output_clocks))
    clock_names = ["ADC", "FPGA"]

    clk.set_requested_clocks(vcxo, output_clocks, clock_names)

    clk.solve()
    o = clk.get_config()

    assert isinstance(o["vcxo"], (int, float))
    assert 100000000 <= o["vcxo"] <= 150000000


def test_ltc6953_arb_source_input_ref():
    """Test LTC6953 with arb_source input_ref."""
    input_ref = adijif.types.arb_source(
        "ref_in", a_min=1000000000, a_max=2000000000, b_min=1, b_max=1
    )

    clk = adijif.ltc6953(solver="CPLEX")

    output_clocks = [1e9, 500e6]
    output_clocks = list(map(int, output_clocks))
    clock_names = ["ADC", "FPGA"]

    clk.set_requested_clocks(input_ref, output_clocks, clock_names)

    clk.solve()

    o = clk.get_config()

    assert isinstance(o["input_ref"], (int, float))
    assert 1000000000 <= o["input_ref"] <= 2000000000


def test_system_with_arb_source_vcxo():
    """Test system class with arb_source vcxo."""
    vcxo = adijif.types.arb_source(
        "vcxo", a_min=90000000, a_max=110000000, b_min=1, b_max=1
    )

    sys = adijif.system("ad9680", "hmc7044", "xilinx", vcxo, solver="CPLEX")
    sys.fpga.setup_by_dev_kit_name("zc706")
    sys.fpga.force_qpll = True

    sys.converter.sample_clock = 1e9
    sys.converter.datapath_decimation = 1
    sys.converter.L = 4
    sys.converter.M = 2
    sys.converter.N = 14
    sys.converter.Np = 16
    sys.converter.K = 32
    sys.converter.F = 1
    sys.converter.HD = 0

    cfg = sys.solve()

    assert "vcxo" in cfg["clock"]
    assert isinstance(cfg["clock"]["vcxo"], (int, float))
    assert 90000000 <= cfg["clock"]["vcxo"] <= 110000000


def test_system_with_range_vcxo():
    """Test system class with range vcxo."""
    vcxo = adijif.types.range(90000000, 110000000, 5000000, "vcxo")

    sys = adijif.system("ad9680", "hmc7044", "xilinx", vcxo, solver="CPLEX")
    sys.fpga.setup_by_dev_kit_name("zc706")
    sys.fpga.force_qpll = True

    sys.converter.sample_clock = 1e9
    sys.converter.datapath_decimation = 1
    sys.converter.L = 4
    sys.converter.M = 2
    sys.converter.N = 14
    sys.converter.Np = 16
    sys.converter.K = 32
    sys.converter.F = 1
    sys.converter.HD = 0

    cfg = sys.solve()

    assert "vcxo" in cfg["clock"]
    assert isinstance(cfg["clock"]["vcxo"], (int, float))
    assert 90000000 <= cfg["clock"]["vcxo"] <= 110000000


def test_arb_source_with_gekko_error():
    """Test that arb_source with GEKKO raises helpful error."""
    vcxo = adijif.types.arb_source(
        "vcxo", a_min=100000000, a_max=150000000, b_min=1, b_max=1
    )

    msg = r"arb_source type requires CPLEX solver"

    with pytest.raises(Exception, match=msg):
        sys = adijif.system("ad9680", "hmc7044", "xilinx", vcxo, solver="gekko")


def test_ltc6952_object_creation():
    """Test LTC6952 object creation."""
    clk = adijif.ltc6952(solver="CPLEX")
    assert clk is not None
    assert hasattr(clk, "set_requested_clocks")


def test_ltc6953_with_range_input():
    """Test LTC6953 with range input reference."""
    ref_in = adijif.types.range(1000000000, 4500000000, 100000000, "ref_in")

    clk = adijif.ltc6953(solver="CPLEX")

    output_clocks = [2e9, 1e9, 500e6]
    output_clocks = list(map(int, output_clocks))
    clock_names = ["ADC", "FPGA", "SYSREF"]

    clk.set_requested_clocks(ref_in, output_clocks, clock_names)

    clk.solve()

    o = clk.get_config()

    assert o is not None
    assert "out_dividers" in o
    assert "input_ref" in o


@pytest.mark.parametrize("solver", ["gekko", "CPLEX"])
def test_hmc7044_basic(solver):
    """Test HMC7044 basic configuration."""
    skip_solver(solver)

    vcxo = 122.88e6

    clk = adijif.hmc7044(solver=solver)

    output_clocks = [245.76e6, 245.76e6, 7.68e6]
    output_clocks = list(map(int, output_clocks))
    clock_names = ["ADC", "FPGA", "SYSREF"]

    clk.set_requested_clocks(vcxo, output_clocks, clock_names)

    clk.solve()
    o = clk.get_config()

    assert o is not None
    assert "vcxo" in o
    assert o["vcxo"] == vcxo


def test_ad9523_1_vcxo_double_attribute():
    """Test AD9523-1 VCXO doubler attribute setting."""
    clk = adijif.ad9523_1(solver="CPLEX")

    # Test that attribute can be set
    clk.use_vcxo_double = True
    assert clk.use_vcxo_double is True


def test_ad9528_vcxo_double_attribute():
    """Test AD9528 VCXO doubler attribute setting."""
    clk = adijif.ad9528(solver="CPLEX")

    # Test that attribute can be set
    clk.use_vcxo_double = True
    assert clk.use_vcxo_double is True



"""AD9084 high speed MxFE clocking model."""

from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List, Union

from ..solvers import GEKKO, CpoModel, CpoSolveResult  # type: ignore
from .ad9084_dp import ad9084_dp_rx
from .ad9084_draw import ad9084_draw
from .ad9084_util import _load_rx_config_modes, apply_settings
from .ad9084_util import parse_json_config as parse_json_cfg
from .ad9088_dp import ad9088_dp_rx
from .adc import adc
from .converter import converter

# from .ad9081_util import _load_rx_config_modes
# from .dac import dac


class ad9084_core(ad9084_draw, converter, metaclass=ABCMeta):
    """AD9084 high speed MxFE model.

    FIXME: This model supports both direct clock configurations and on-board
    generation

    Once we have the DAC clock the data rates can be directly evaluated into
    each JESD framer:

    rx_baseband_sample_rate = (dac_clock / L) / datapath_decimation
    tx_baseband_sample_rate = dac_clock / datapath_interpolation

    """

    device_clock_available = None  # FIXME
    device_clock_ranges = None  # FIXME

    model: Union[GEKKO, CpoModel] = None

    name = "AD9084"

    # # Integrated PLL constants
    # l_available = [1, 2, 3, 4]
    # l = 1  # pylint:  disable=E741
    # m_vco_available = [5, 7, 8, 11]  # 8 is nominal
    # m_vco = 8
    # n_vco_available = [*range(2, 50 + 1)]
    # n_vco = 2
    # r_available = [1, 2, 3, 4]
    # r = 1
    # d_available = [1, 2, 3, 4]
    # d = 1
    # # Integrated PLL limits
    # pfd_min = 25e6
    # pfd_max = 750e6
    # vco_min = 6e9
    # vco_max = 12e9

    # JESD parameters
    available_jesd_modes = ["jesd204b", "jesd204c"]
    M_available = [1, 2, 3, 4, 6, 8, 12, 16]
    L_available = [1, 2, 3, 4, 6, 8, 12]
    N_available = [12, 16]
    Np_available = [8, 12, 16, 24]
    F_available = [1, 2, 3, 4, 6, 8, 12, 16, 24, 32]
    S_available = [1, 2, 3, 4, 6, 8, 12, 16]
    # FIXME
    # K_available = [4, 8, 12, 16, 20, 24, 28, 32]
    K_available = [16, 32, 64, 128, 256]
    CS_available = [0, 1, 2, 3]
    CF_available = [0]
    # FIXME

    # FIXME: These are not known yet
    # Clocking constraints
    # clocking_option_available = ["integrated_pll", "direct", "external"]
    clocking_option_available = ["direct"]
    _clocking_option = "direct"
    bit_clock_min_available = {"jesd204b": 1.5e9, "jesd204c": 1e9}  # FIXME: Wrong
    bit_clock_max_available = {"jesd204b": 15.5e9, "jesd204c": 28.2e9}

    config = {}  # type: ignore

    device_clock_max = 12e9
    _model_type = "adc"

    def _check_valid_internal_configuration(self) -> None:
        # FIXME
        pass

    def apply_profile_settings(
        self, profile_json: str, bypass_version_check: bool = False
    ) -> None:
        """Parse Apollo profiles and apply settings to the model.

        Args:
            profile_json (str): Path to the profile JSON file.
            bypass_version_check (bool): Bypass the version check for profile
        """
        settings = parse_json_cfg(profile_json, bypass_version_check)
        apply_settings(self, settings)

    def get_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect internal converter configuration and output clock definitions
        leading to connected devices (clock chips, FPGAs)

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration
        """
        if solution:
            self.solution = solution
            self._saved_solution = solution

        if self.clocking_option == "integrated_pll":
            pll_config: Dict = {
                "m_vco": self._get_val(self.config["m_vco"]),
                "n_vco": self._get_val(self.config["n_vco"]),
                "r": self._get_val(self.config["r"]),
                "d": self._get_val(self.config["d"]),
            }
            return {"clocking_option": self.clocking_option, "pll_config": pll_config}
        else:
            return {"clocking_option": self.clocking_option}

    def get_required_clock_names(self) -> List[str]:
        """Get list of strings of names of requested clocks.

        This list of names is for the clocks defined by get_required_clocks

        Returns:
            List[str]: List of strings of clock names in order
        """
        name = "AD9084" if "9084" in self.name else "AD9088"
        return [f"{name}_ref_clk", f"{name}_sysref"]

    @property
    @abstractmethod
    def _converter_clock_config(self) -> None:
        """Define source clocking relation based on ADC, DAC, or both.

        Raises:
            NotImplementedError: Method not implemented
        """
        raise NotImplementedError

    def _pll_config(self, rxtx: bool = False) -> Dict:
        self._converter_clock_config()  # type: ignore

        self.config["m_vco"] = self._convert_input([5, 7, 8, 11], "m_vco")
        self.config["n_vco"] = self._convert_input([*range(2, 51)], "n_vco")
        self.config["r"] = self._convert_input([1, 2, 3, 4], "r")
        self.config["d"] = self._convert_input([1, 2, 3, 4], "d")

        self.config["ref_clk"] = self._add_intermediate(
            self.config["converter_clk"]
            * self.config["d"]
            * self.config["r"]
            / (self.config["m_vco"] * self.config["n_vco"])
        )
        # if self.solver == "gekko":
        #     self.config["ref_clk"] = self.model.Var(
        #         integer=True,
        #         lb=1e6,
        #         ub=self.device_clock_max,
        #         value=self.device_clock_max,
        #     )
        # elif self.solver == "CPLEX":
        #     # self.config["ref_clk"] = integer_var(
        #     #     int(1e6), int(self.device_clock_max), "ref_clk"
        #     # )
        #     self.config["ref_clk"] = (
        #         self.config["converter_clk"]
        #         * self.config["d"]
        #         * self.config["r"]
        #         / (self.config["m_vco"] * self.config["n_vco"])
        #     )
        # else:
        #     raise Exception("Unknown solver")

        self.config["vco"] = self._add_intermediate(
            self.config["ref_clk"]
            * self.config["m_vco"]
            * self.config["n_vco"]
            / self.config["r"],
        )

        # if self.solver == "gekko":
        #     self.config["vco"] = self.model.Intermediate(
        #         self.config["ref_clk"]
        #         * self.config["m_vco"]
        #         * self.config["n_vco"]
        #         / self.config["r"],
        #     )
        # elif self.solver == "CPLEX":
        #     self.config["vco"] = (
        #         self.config["ref_clk"]
        #         * self.config["m_vco"]
        #         * self.config["n_vco"]
        #         / self.config["r"]
        #     )
        # else:
        #     raise Exception("Unknown solver: %s" % self.solver)

        self._add_equation(
            [
                self.config["vco"] >= self.vco_min,
                self.config["vco"] <= self.vco_max,
                self.config["ref_clk"] / self.config["r"] <= self.pfd_max,
                self.config["ref_clk"] / self.config["r"] >= self.pfd_min,
                # self.config["converter_clk"] <= self.device_clock_max,
                self.config["converter_clk"]
                >= (
                    self.converter_clock_min
                    if not rxtx
                    else self.dac.converter_clock_min  # type: ignore
                ),
                self.config["converter_clk"]
                <= (
                    self.converter_clock_max
                    if not rxtx
                    else self.dac.converter_clock_max  # type: ignore
                ),
            ]
        )

        return self.config["ref_clk"]

    def get_required_clocks(self) -> List:
        """Generate list required clocks.

        For AD9084 this will contain [converter clock, sysref requirement SOS]

        Returns:
            List: List of solver variables, equations, and constants
        """
        # SYSREF
        self.config = {}
        self.config["lmfc_divisor_sysref"] = self._convert_input(
            [*range(1, 21)], "lmfc_divisor_sysref"
        )

        if self.solver == "gekko":
            self.config["sysref"] = self.model.Intermediate(
                self.multiframe_clock  # type: ignore
                / (
                    self.config["lmfc_divisor_sysref"]
                    * self.config["lmfc_divisor_sysref"]
                )
            )
        elif self.solver == "CPLEX":
            self.config["sysref"] = self.multiframe_clock / (
                self.config["lmfc_divisor_sysref"] * self.config["lmfc_divisor_sysref"]
            )

        # Device Clocking
        if self.clocking_option == "direct":
            clk = self.sample_clock * self.datapath.decimation_overall
        elif self.clocking_option == "external":
            return [[], self.config["sysref"]]
        else:
            clk = self._pll_config()  # type: ignore

        # Objectives
        # self.model.Obj(self.config["sysref"])  # This breaks many searches
        # self.model.Obj(-1*self.config["lmfc_divisor_sysref"])

        return [clk, self.config["sysref"]]


class ad9084_rx(adc, ad9084_core):
    """AD9084 Receive model."""

    converter_type = "adc"
    name = "AD9084_RX"

    converter_clock_min = 1e9  # FIXME
    converter_clock_max = 20e9

    sample_clock_min = 312.5e6 / 16  # FIXME
    sample_clock_max = 20e9

    quick_configuration_modes = _load_rx_config_modes(part="AD9084")

    datapath = ad9084_dp_rx()

    decimation_available = [
        cddc * fddc for cddc in [1, 2, 3, 4, 6, 12] for fddc in [1, 2, 4, 8, 16, 32, 64]
    ]

    @property
    def decimation(self) -> int:
        """Decimation factor. This is the product of the coarse and fine decimation."""
        return self.datapath.decimation_overall

    @decimation.setter
    def decimation(self, value: int) -> None:
        raise Exception(
            "Decimation is not writable and should be set by the properties\n"
            + " datapath.cddc_decimations and datapath.fddc_decimations"
        )

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize AD9084 clocking model for RX.

        This is a common class used to handle RX constraints
        together.

        Args:
            *args (Any): Pass through arguments
            **kwargs (Any): Pass through keyword arguments
        """
        self.sample_clock = int(2.5e9) if "AD9084" in self.name else int(1e9)
        N = 1 if "9084" in self.name else 2
        self.datapath.cddc_decimations = [4] * 4 * N
        self.datapath.fddc_decimations = [2] * 8 * N
        self.datapath.fddc_enabled = [True] * 8 * N
        mode = "47" if "9084" in self.name else "45"

        self.set_quick_configuration_mode(mode, "jesd204c")

        super().__init__(*args, **kwargs)
        self._init_diagram()

    def _converter_clock_config(self) -> None:
        """RX specific configuration of internall PLL config.

        This method will update the config struct to include
        the RX clocking constraints

        Raises:
            Exception: If solver is not valid
        """
        adc_clk = self.decimation * self.sample_clock
        # FIXME: Not sure if this divider is here anymore
        # self.config["l"] = self._convert_input([1, 2, 3, 4], "l")
        self.config["l"] = self._convert_input([1], "l")
        self.config["adc_clk"] = self._convert_input(adc_clk)

        if self.solver == "gekko":
            self.config["converter_clk"] = self.model.Intermediate(
                self.config["adc_clk"] * self.config["l"]
            )
        elif self.solver == "CPLEX":
            self.config["converter_clk"] = self.config["adc_clk"] * self.config["l"]
        else:
            raise Exception(f"Unknown solver {self.solver}")

    def _check_valid_internal_configuration(self) -> None:
        ...
        # mode = self._check_valid_jesd_mode()
        # cfg = self.quick_configuration_modes[self.jesd_class][mode]

        # Check decimation is valid
        # if isinstance(self.decimation, int) or isinstance(self.decimation, float):
        #     found = False
        #     for dec in cfg["decimations"]:
        #         found = found or dec["coarse"] * dec["fine"] == self.decimation
        #     assert (
        #         found
        #     ), f"Decimation {self.decimation} not valid for current JESD mode"
        # elif self.decimation == "auto":
        #     for dec in cfg["decimations"]:
        #         dec = dec["coarse"] * dec["fine"]
        #         # Check
        #         cc = dec * self.sample_clock
        #         # if dec == 64:
        #         #     print("dec", dec, cc, cfg["coarse"], cfg["fine"])
        #         if cc <= self.converter_clock_max and cc >= self.converter_clock_min:
        #             self.decimation = dec
        #             print("Decimation automatically determined:", dec)
        #             return
        #     raise Exception("No valid decimation found")
        # else:
        #     raise Exception("Decimation not valid")


class ad9088_rx(ad9084_rx):
    """AD9088 Receive model."""

    converter_type = "adc"
    name = "AD9088_RX"

    converter_clock_min = 2.9e9
    converter_clock_max = 8e9

    sample_clock_min = 2.9e9 / (6 * 24)  # with max decimation
    sample_clock_max = 8e9

    datapath = ad9088_dp_rx()

    quick_configuration_modes = _load_rx_config_modes(part="AD9088")


# class ad9084_tx(dac, ad9084_core):
#     """AD9084 Transmit model."""

#     _model_type = "dac"
#     name = "AD9084_TX"

#     converter_clock_min = 2.9e9
#     converter_clock_max = 12e9

#     sample_clock_min = 2.9e9 / (6 * 24)  # with max interpolation
#     sample_clock_max = 12e9

#     quick_configuration_modes = _load_tx_config_modes()

#     datapath = ad9084_dp_tx()
#     interpolation_available = [
#         1,
#         2,
#         3,
#         4,
#         6,
#         8,
#         9,
#         12,
#         16,
#         18,
#         24,
#         32,
#         36,
#         48,
#         64,
#         72,
#         96,
#         144,
#     ]
#     interpolation = 1

#     def __init__(
#         self, model: Union[GEKKO, CpoModel] = None, solver: str = None
#     ) -> None:
#         """Initialize AD9084 clocking model for TX.

#         This is a common class used to handle TX constraints
#         together.

#         Args:
#             model (GEKKO,CpoModel): Solver model
#             solver (str): Solver name (gekko or CPLEX)
#         """
#         if solver:
#             self.solver = solver
#         if model:
#             self.model = model
#         self.set_quick_configuration_mode("0", "jesd204c")

#     def _converter_clock_config(self) -> None:
#         """TX specific configuration of internall PLL config.

#         This method will update the config struct to include
#         the TX clocking constraints

#         Raises:
#             Exception: If solver is not valid
#         """
#         dac_clk = self.interpolation * self.sample_clock
#         self.config["dac_clk"] = self._convert_input(dac_clk)
#         if self.solver == "gekko":
#             self.config["converter_clk"] = self.model.Intermediate(
#                 self.config["dac_clk"]
#             )
#         elif self.solver == "CPLEX":
#             self.config["converter_clk"] = self.config["dac_clk"]
#         else:
#             raise Exception(f"Unknown solver {self.solver}")


# class ad9084(ad9084_core):
#     """AD9084 combined transmit and receive model."""

#     converter_clock_min = ad9084_rx.converter_clock_min
#     converter_clock_max = ad9084_rx.converter_clock_max
#     quick_configuration_modes: Dict[str, Any] = {}
#     _nested = ["adc", "dac"]

#     def __init__(
#         self, model: Union[GEKKO, CpoModel] = None, solver: str = None
#     ) -> None:
#         """Initialize AD9084 clocking model for TX and RX.

#         This is a common class used to handle TX and RX constraints
#         together.

#         Args:
#             model (GEKKO,CpoModel): Solver model
#             solver (str): Solver name (gekko or CPLEX)
#         """
#         if solver:
#             self.solver = solver
#         self.adc = ad9084_rx(model, solver=self.solver)
#         self.dac = ad9084_tx(model, solver=self.solver)
#         self.model = model

#     def validate_config(self) -> None:
#         """Validate device configurations including JESD and clocks of both
#         ADC and DAC.

#         This check only is for static configuration that does not include
#         variables which are solved.
#         """
#         self.adc.validate_config()
#         self.dac.validate_config()

#     def _get_converters(self) -> List[Union[converter, converter]]:
#         return [self.adc, self.dac]

#     def get_required_clock_names(self) -> List[str]:
#         """Get list of strings of names of requested clocks.

#         This list of names is for the clocks defined by get_required_clocks

#         Returns:
#             List[str]: List of strings of clock names in order
#         """
#         clk = (
#             "ad9084_dac_clock"
#             if self.adc.clocking_option == "direct"
#             else "ad9084_pll_ref"
#         )
#         return [clk, "ad9084_adc_sysref", "ad9084_dac_sysref"]

#     def _converter_clock_config(self) -> None:
#         adc_clk = self.adc.decimation * self.adc.sample_clock
#         dac_clk = self.dac.interpolation * self.dac.sample_clock
#         l = dac_clk / adc_clk
#         if np.abs(l - round(l)) > 1e-6:
#             raise Exception("Sample clock ratio is not integer")
#         else:
#             l = int(round(l))
#         if l not in self.adc.l_available:
#             raise Exception(
#                 f"ADC clock must be DAC clock/L where L={self.adc.l_available}."
#                 + f" Got {l} ({dac_clk}/{adc_clk})"
#             )

#         self.config["dac_clk"] = self._convert_input(dac_clk)
#         self.config["adc_clk"] = self._convert_input(adc_clk)
#         if self.solver == "gekko":
#             self.config["converter_clk"] = self.model.Intermediate(
#                 self.config["dac_clk"]
#             )
#         elif self.solver == "CPLEX":
#             self.config["converter_clk"] = self.config["dac_clk"]
#         else:
#             raise Exception(f"Unknown solver {self.solver}")

#     def get_required_clocks(self) -> List:
#         """Generate list required clocks.

#         For AD9084 this will contain [converter clock, sysref requirement SOS]

#         Returns:
#             List: List of solver variables, equations, and constants

#         Raises:
#             Exception: If direct clocking is used. Not yet implemented
#         """
#         # SYSREF
#         self.config = {}
#         self.config["adc_lmfc_divisor_sysref"] = self._convert_input(
#             [*range(1, 21)], "adc_lmfc_divisor_sysref"
#         )
#         self.config["dac_lmfc_divisor_sysref"] = self._convert_input(
#             [*range(1, 21)], "dac_lmfc_divisor_sysref"
#         )

#         if self.solver == "gekko":
#             self.config["sysref_adc"] = self.model.Intermediate(
#                 self.adc.multiframe_clock
#                 / (
#                     self.config["adc_lmfc_divisor_sysref"]
#                     * self.config["adc_lmfc_divisor_sysref"]
#                 )
#             )
#             self.config["sysref_dac"] = self.model.Intermediate(
#                 self.dac.multiframe_clock
#                 / (
#                     self.config["dac_lmfc_divisor_sysref"]
#                     * self.config["dac_lmfc_divisor_sysref"]
#                 )
#             )
#         elif self.solver == "CPLEX":
#             self.config["sysref_adc"] = self.adc.multiframe_clock / (
#                 self.config["adc_lmfc_divisor_sysref"]
#                 * self.config["adc_lmfc_divisor_sysref"]
#             )
#             self.config["sysref_dac"] = self.dac.multiframe_clock / (
#                 self.config["dac_lmfc_divisor_sysref"]
#                 * self.config["dac_lmfc_divisor_sysref"]
#             )
#         else:
#             raise Exception(f"Unknown solver {self.solver}")

#         # Device Clocking
#         if self.clocking_option == "direct":
#             raise Exception("Not implemented yet")
#             # adc_clk = self.sample_clock * self.datapath_decimation
#         else:
#             clk = self._pll_config(rxtx=True)

#         # Objectives
#         # self.model.Obj(self.config["sysref"])  # This breaks many searches
#         # self.model.Obj(-1*self.config["lmfc_divisor_sysref"])

#         return [clk, self.config["sysref_adc"], self.config["sysref_dac"]]



"""Converter base meta class for all converter clocking models."""

from abc import ABCMeta, abstractmethod
from typing import Dict, List, Union

from ..common import core
from ..draw import Layout, Node
from ..gekko_trans import gekko_translation
from ..jesd import jesd


class converter(core, jesd, gekko_translation, metaclass=ABCMeta):
    """Converter base meta class used to enforce all converter classes.

    This class should be inherited from for all converters: ADCs, DACs,
    and transceivers.

    """

    """Nested device with multiple dependent converters. Includes MxFE and
    transceivers"""
    _nested = False

    """DSP data path of device"""
    datapath = None

    """Show rates on drawing diagrams."""
    show_rates = True

    config: Dict = {}
    _jesd_params_to_skip = [
        "E",
        "global_index",
    ]  # Params in table not to set
    _jesd_params_to_skip_check = [
        "DualLink",
        "E",
        "decimations",
        "global_index",
    ]

    def draw_dac(
        self, clocks: Dict, lo: Layout = None, clock_chip_node: Node = None
    ) -> str:
        """Generic Draw converter DAC model.

        Args:
            clocks (Dict): Clocking configuration
            lo (Layout): Layout object to draw on
            clock_chip_node (Node): Clock chip node to add to. Defaults to None.

        Returns:
            str: Path to image file

        Raises:
            Exception: If no solution is saved
        """
        system_draw = lo is not None
        name = self.name.lower()

        if not system_draw:
            lo = Layout(f"{name} Example")
            lo.show_rates = self.show_rates
        else:
            assert isinstance(lo, Layout), "lo must be a Layout object"

        ic_node = Node(self.name)
        lo.add_node(ic_node)

        # rate = clocks[f"{name}_ref_clk"]
        # Find key with ending
        ref_clk_name = None
        for key in clocks.keys():
            if key.lower().endswith(f"{name.lower()}_ref_clk"):
                ref_clk_name = key
                break
        if ref_clk_name is None:
            raise Exception(
                f"No clock found for {name}_ref_clk\n.Options: {clocks.keys()}"
            )

        sysref_clk_name = None
        for key in clocks.keys():
            if key.lower().endswith(f"{name.lower()}_sysref"):
                sysref_clk_name = key
                break
        if sysref_clk_name is None:
            raise Exception(
                f"No clock found for {name}_sysref\n.Options: {clocks.keys()}"
            )

        if not system_draw:
            ref_in = Node("REF_IN", ntype="input")
            lo.add_node(ref_in)
        else:
            to_node = lo.get_node(ref_clk_name)
            from_node = lo.get_connection(to=to_node.name)
            assert from_node, "No connection found"
            assert isinstance(from_node, list), "Connection must be a list"
            assert len(from_node) == 1, "Only one connection allowed"
            ref_in = from_node[0]["from"]
            # Remove to_node since it is not needed
            lo.remove_node(to_node.name)

        rate = clocks[ref_clk_name]

        lo.add_connection({"from": ref_in, "to": ic_node, "rate": rate})

        # Add deframer
        jesd204_deframer = Node("JESD204 Deframer", ntype="jesd204deframer")
        ic_node.add_child(jesd204_deframer)

        # Add DAC
        dac = Node("DAC", ntype="dac")
        ic_node.add_child(dac)
        ic_node.add_connection({"from": jesd204_deframer, "to": dac})

        # SYSREF
        if not system_draw:
            sysref_in = Node("SYSREF_IN", ntype="input")
            lo.add_connection(
                {
                    "from": sysref_in,
                    "to": jesd204_deframer,
                    # "rate": clocks[f"{name}_sysref"],
                    "rate": clocks[sysref_clk_name],
                }
            )
        else:
            # to_node = lo.get_node(f"{name}_sysref")
            to_node = lo.get_node(sysref_clk_name)
            from_node = lo.get_connection(to=to_node.name)
            assert from_node, "No connection found"
            assert isinstance(from_node, list), "Connection must be a list"
            assert len(from_node) == 1, "Only one connection allowed"
            sysref_in = from_node[0]["from"]
            lo.remove_node(to_node.name)

            lo.add_connection(
                {
                    "from": sysref_in,
                    "to": jesd204_deframer,
                    # "rate": clocks[f"{name}_sysref"],
                    "rate": clocks[sysref_clk_name],
                }
            )

        # WIP Add remote framer
        jesd204_framer = Node("JESD204 Framer", ntype="framer")
        lo.add_node(jesd204_framer)

        # Add connect for each lane
        for _ in range(self.L):
            lane_rate = self.bit_clock
            lo.add_connection(
                {"from": jesd204_framer, "to": jesd204_deframer, "rate": lane_rate}
            )

        if not system_draw:
            return lo.draw()

    def draw_adc(
        self, clocks: Dict, lo: Layout = None, clock_chip_node: Node = None
    ) -> str:
        """Generic Draw converter ADC model.

        Args:
            clocks (Dict): Clocking configuration
            lo (Layout): Layout object to draw on
            clock_chip_node (Node): Clock chip node to add to. Defaults to None.

        Returns:
            str: Path to image file

        Raises:
            Exception: If no solution is saved
        """
        system_draw = lo is not None
        name = self.name.lower()

        if not system_draw:
            lo = Layout(f"{name} Example")
            lo.show_rates = self.show_rates
        else:
            assert isinstance(lo, Layout), "lo must be a Layout object"

        ic_node = Node(self.name)
        lo.add_node(ic_node)

        # rate = clocks[f"{name}_ref_clk"]
        # Find key with ending
        ref_clk_name = None
        for key in clocks.keys():
            if key.lower().endswith(f"{name.lower()}_ref_clk"):
                ref_clk_name = key
                break
        if ref_clk_name is None:
            raise Exception(
                f"No clock found for {name}_ref_clk\n.Options: {clocks.keys()}"
            )

        sysref_clk_name = None
        for key in clocks.keys():
            if key.lower().endswith(f"{name.lower()}_sysref"):
                sysref_clk_name = key
                break
        if sysref_clk_name is None:
            raise Exception(
                f"No clock found for {name}_sysref\n.Options: {clocks.keys()}"
            )

        if not system_draw:
            ref_in = Node("REF_IN", ntype="input")
            lo.add_node(ref_in)
        else:
            to_node = lo.get_node(ref_clk_name)
            from_node = lo.get_connection(to=to_node.name)
            assert from_node, "No connection found"
            assert isinstance(from_node, list), "Connection must be a list"
            assert len(from_node) == 1, "Only one connection allowed"
            ref_in = from_node[0]["from"]
            # Remove to_node since it is not needed
            lo.remove_node(to_node.name)

        rate = clocks[ref_clk_name]

        lo.add_connection({"from": ref_in, "to": ic_node, "rate": rate})

        # Add ADC
        adc = Node("ADC", ntype="adc")
        ic_node.add_child(adc)

        # Add framer
        jesd204_framer = Node("JESD204 Framer", ntype="jesd204framer")
        ic_node.add_child(jesd204_framer)
        ic_node.add_connection({"from": adc, "to": jesd204_framer})

        # SYSREF
        if not system_draw:
            sysref_in = Node("SYSREF_IN", ntype="input")
            lo.add_connection(
                {
                    "from": sysref_in,
                    "to": jesd204_framer,
                    # "rate": clocks[f"{name}_sysref"],
                    "rate": clocks[sysref_clk_name],
                }
            )
        else:
            # to_node = lo.get_node(f"{name}_sysref")
            to_node = lo.get_node(sysref_clk_name)
            from_node = lo.get_connection(to=to_node.name)
            assert from_node, "No connection found"
            assert isinstance(from_node, list), "Connection must be a list"
            assert len(from_node) == 1, "Only one connection allowed"
            sysref_in = from_node[0]["from"]
            lo.remove_node(to_node.name)

            lo.add_connection(
                {
                    "from": sysref_in,
                    "to": jesd204_framer,
                    # "rate": clocks[f"{name}_sysref"],
                    "rate": clocks[sysref_clk_name],
                }
            )

        # WIP Add remote deframer
        jesd204_deframer = Node("JESD204 Deframer", ntype="deframer")
        lo.add_node(jesd204_deframer)

        # Add connect for each lane
        for _ in range(self.L):
            lane_rate = self.bit_clock
            lo.add_connection(
                {"from": jesd204_framer, "to": jesd204_deframer, "rate": lane_rate}
            )

        if not system_draw:
            return lo.draw()

    def draw(
        self, clocks: Dict, lo: Layout = None, clock_chip_node: Node = None
    ) -> str:
        """Generic Draw converter model.

        Args:
            clocks (Dict): Clocking configuration
            lo (Layout): Layout object to draw on
            clock_chip_node (Node): Clock chip node to add to. Defaults to None.

        Returns:
            str: Path to image file

        Raises:
            Exception: If no solution is saved
        """
        if self.converter_type.lower() == "adc":
            return self.draw_adc(clocks, lo, clock_chip_node)
        elif self.converter_type.lower() == "dac":
            return self.draw_dac(clocks, lo, clock_chip_node)
        else:
            raise Exception("Unknown converter type")

    def validate_config(self) -> None:
        """Validate device configuration including JESD and clocks.

        This check only is for static configuration that does not include
        variables which are solved.
        """
        self._check_valid_jesd_mode()
        self._check_valid_internal_configuration()  # type: ignore
        self.validate_clocks()

    def set_quick_configuration_mode(
        self, mode: Union[str, int], jesd_class: str = "jesd204b"
    ) -> None:
        """Set JESD configuration based on preset mode table.

        Args:
            mode (str): Integer of desired mode.
            jesd_class (str): JESD class to use. Default is jesd204b.

        Raises:
            Exception: Invalid mode selected
        """
        smode = str(mode)
        if jesd_class not in self.available_jesd_modes:
            raise Exception(
                f"{jesd_class} not available for JESD class for {self.name}"
            )

        if smode not in self.quick_configuration_modes[jesd_class].keys():
            raise Exception(f"Mode {smode} not among configurations for {jesd_class}")

        self.jesd_class = jesd_class

        cfg = self.quick_configuration_modes[jesd_class][smode]
        for jparam in cfg:
            if jparam in self._jesd_params_to_skip:
                continue
            setattr(self, jparam, cfg[jparam])

    def _check_valid_jesd_mode(self) -> str:
        """Verify current JESD configuration for part is valid.

        Raises:
            Exception: Invalid JESD configuration

        Returns:
            str: Current JESD mode
        """
        # Check to make sure JESD clocks in range
        self._check_jesd_config()
        # Pull current mode
        k = next(iter(self.quick_configuration_modes[self.jesd_class]))
        attrs = self.quick_configuration_modes[self.jesd_class][k].keys()

        current_config = {}
        for attr in attrs:
            if attr in self._jesd_params_to_skip_check:
                continue
            current_config[attr] = getattr(self, attr)

        # Check mode in supported modes
        for mode in self.quick_configuration_modes[self.jesd_class].keys():
            cmode = self.quick_configuration_modes[self.jesd_class][mode].copy()
            for k in self._jesd_params_to_skip_check:
                if hasattr(cmode, k) or k in cmode:
                    del cmode[k]
                if hasattr(current_config, k) or k in current_config:
                    del current_config[k]
            if current_config == cmode:
                return mode
        raise Exception(f"Invalid JESD configuration for {self.name}\n{current_config}")

    def get_current_jesd_mode_settings(self) -> Dict:
        """Get current JESD mode settings.

        Returns:
            Dict: Current JESD mode settings
            str: Current JESD mode
        """
        k = next(iter(self.quick_configuration_modes[self.jesd_class]))
        attrs = self.quick_configuration_modes[self.jesd_class][k].keys()
        current_config = {}
        for attr in attrs:
            if attr in self._jesd_params_to_skip_check:
                continue
            current_config[attr] = getattr(self, attr)
        return current_config

    @property
    @abstractmethod
    def converter_type(self) -> str:
        """Type of converter. ADC or DAC.

        Returns:
            str: Type of converter
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def clocking_option_available(self) -> List[str]:
        """Supported clocking modes.

        Returns:
            List: List of string of clocking modes
        """
        raise NotImplementedError  # pragma: no cover

    _clocking_option = "direct"

    @property
    def clocking_option(self) -> str:
        """Get clocking mode for device.

        Returns:
            str: Clocking mode for device (integrated_pll, direct)
        """
        return self._clocking_option

    @clocking_option.setter
    def clocking_option(self, value: str) -> None:
        """Set clocking mode for device.

        Args:
            value (str): Clocking mode for device (integrated_pll, direct)

        Raises:
            Exception: clocking_option not supported by device
        """
        if value not in self.clocking_option_available:
            raise Exception("clocking_option not available for device")
        self._clocking_option = value

    @property
    @abstractmethod
    def quick_configuration_modes(self) -> Dict:
        """Supported JESD mode table.

        Returns:
            Dict: Dictionary of supported modes
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def _check_valid_internal_configuration(self) -> None:
        """Check current device mode is valid."""
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def converter_clock_min(self) -> Union[int, float]:
        """Minimum rate of data converter.

        Returns:
            Union[int, float]: converter min rate
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def converter_clock_max(self) -> Union[int, float]:
        """Maximum rate of data converter.

        Returns:
            Union[int, float]: converter max rate
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of supported by device.

        Must be a string

        Raises:
            NotImplementedError: If child classes do not implement method/property
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def device_clock_available(self) -> None:
        """Allowable K settings for device.

        Must be a list ints

        Raises:
            NotImplementedError: If child classes do not implement method/property
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def device_clock_ranges(self) -> None:
        """Allowable ranges for device clocks based on config.

        This is only used for brute-force implementations and not solver based

        Raises:
            NotImplementedError: If child classes do not implement method/property
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def get_required_clocks(self) -> List:
        """Get list of strings of names of requested clocks.

        This list of names is for the clocks defined by get_required_clocks

        Raises:
            NotImplementedError: If child classes do not implement method/property
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def get_required_clock_names(self) -> List[str]:
        """Generate list of required clock names.

        This is a list of strings

        Raises:
            NotImplementedError: If child classes do not implement method/property
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def get_config(self) -> Dict:
        """Extract configuration from solver solution.

        Return dictionary of clocking config for converter

        Raises:
            NotImplementedError: If child classes do not implement method/property
        """
        raise NotImplementedError  # pragma: no cover

    def _get_converters(self) -> Union["converter", List["converter"]]:
        return self

    def __str__(self) -> str:
        """Get string description of converter object.

        Returns:
            str: Description string
        """
        return f"{self.name} data converter model"
